import numpy as np
from skimage.segmentation import watershed
from skimage.feature import peak_local_max, blob_log, blob_dog
from skimage.filters import gaussian
from typing import Union, Callable
import napari
import json
from iterseg.predict import process_chunks, predict_chunk_feature_map, load_unet
from iterseg import watershed as ws
import zarr
import os
from napari.qt import thread_worker
from scipy import ndimage as ndi
import pathlib
import os


# ------------------------
# Affinity U-net Watershed
# ------------------------

def affinity_unet_watershed(
        napari_viewer, 
        input_volume_layer: napari.layers.Image,
        save_dir: Union[str, None] = None,
        name: str = 'my-segmentation',
        unet_or_config_file: Union[str, None] = None,
        layer_reference: Union[str, None] = None,
        chunk_size: Union[tuple, None] = (10, 256, 256),
        margin: Union[tuple, None] = (1, 64, 64),
        debug: bool =False
    ):
    '''
    Segment a 3D image or stack of 3D images using a affinity-unet-watershed
    algorithm. The image is passed to a U-Net, which outputs 5 feature maps:
    (1) z-axis affinities, (2) y-axis affinities, (3) x-axis affinities, 
    (4) foreground mask, (5) centre point predicition. The feature maps are
    used in a modified watershed algorithms. The centre point prediciton is
    used to find object centres and the foreground mask is used to determine
    the pixels that need to be labeled. The affinities in each axis demarkate
    the boundaries between objects and the waterhshed uses these to find the 
    edges of objects. 

    Parameters
    ----------
    napari_viewer: napari.Viewer
        The viewer into which the resultant segmentation will be placed.
    input_volume_layer: napari.layers.Image
        The viewer layer with the image to be processed.
    save_dir: str or None (None)
        The directory into which to save the outputted segmentation. 
    name: str ('labels-prediction')
        The name with which to save the data. Also the name of the 
        resultant segmentation layer that will be displayed in the 
        viewer. 
    config_file: str or None (None)
        Path to a json file with any additional parameters. I'm sure
        There is  a more elegant way to do this through the GUI, 
        but I don't have the time to work on this for now. 
    layer_reference: str or None (None)
        If you need information from one of the layers, this is the
        layer to get it from. This is fed into the custom config function. 
        This function can get a unet from layer metadata. 
    chunk_size: tuple (10, 256, 256)
         This tells you how big the chunks to be processed are
    margin: tuple (1, 64, 64)
        This tells you how much overlap between chunks is used
    '''
    segmentation_wrapper(affinity_watershed_for_chunks, affinity_watershed_prep_config, 
                         napari_viewer, input_volume_layer, save_dir, name, unet_or_config_file,
                         layer_reference, chunk_size, margin, debug)



# AUW - Config
# ------------

def affinity_watershed_prep_config(
        input_volume_layer,
        unet_or_config_file, 
        reference_layer
        ):
    # No config or unet path provided
    # -------------------------------
    if unet_or_config_file is None:
        unet = None
        affinities_extent = 1
    # Pathlib object though magicgui
    # ------------------------------
    if isinstance(unet_or_config_file, pathlib.PurePath):
        unet_or_config_file = str(unet_or_config_file)
    if isinstance(unet_or_config_file, str):
        affinities_extent = 1
        # Path provided
        # -------------
        if unet_or_config_file.endswith('.json'):
            config = read_config_json(unet_or_config_file)
            if config.get('unet') is None:
                unet = None
            if config.get('affinities_extent') is None:
                affinities_extent = 1
            if unet == 'labels layer':
                unet = reference_layer.metadata['unet']
            if unet == 'default':
                unet = None
        elif unet_or_config_file.endswith('.pt') or unet_or_config_file.endswith('.pth'):
            unet = unet_or_config_file
        # if a path is provided, 
    if unet is not None:
        if isinstance(unet, str):
            m = 'There was not file at the provided location: {unet}\nMake sure a pytorch unet lives here...'
            assert os.path.exists(unet), m
            assert unet.endswith('.pt') or unet.endswith('.pth')
        else:
            raise ValueError('Please provide a valid path to a pytorch unet - must end with .pt or .pth')
    
    # Output volume
    # -------------
    num_pred_channels = 3 * affinities_extent + 2
    output_volume = a_w_output_volume(input_volume_layer.data, num_pred_channels)

    # Load unet
    # ---------
    unet = load_unet(unet)
    
    # key word arguments for segmentation
    # -----------------------------------
    num_pred_channels = 3 * affinities_extent + 2
    new_config = {
        'unet': unet, 
        'output_volume' : output_volume, 
    }
    return new_config


def a_w_output_volume(data, num_pred_channels, **kwargs):
    output_volume = np.zeros((num_pred_channels,) + data.shape[-3:], dtype=np.float32) 
    return output_volume



# AUW - Processing function
# -------------------------

def affinity_watershed_for_chunks(
        input_volume, 
        current_output, 
        chunk_size, 
        margin,
        # specific
        unet=None,
        output_volume=None,
        **kwargs
        ):
    '''
    Function that inserts a the results of an affinity watershed into 
    current output

    Parameters
    ----------
    input_volume: array 
        This is the image to process
    output_volume: array
        This is the array that is used for intermediate processing
        steps (e.g., unet-derived feature map, LoG processed image). 
        In this case, it has five channels for the output of
        the affinity unet (z-1, y-1, x-1, mask, centre-points). 
    chunk_size: tuple
        This tells you how big the chunks to be processed are
    margin: tuple
        This tells you how much overlap between chunks is used
    current_output: array
        This is the 3D array that stores the labels for the particular
        slice. The labels are then written into the labels layer data, 
        which is then displayed on the viewer. 
    unet: str or None
        Path to the unet to load to process the data. This will come
        from the config file. If None, the default unet is used. 
    use_default_unet: bool
        Do you want to get a mask from the default unet (which has been
        trained to find platelets). 
    '''
    config = {'unet' : unet, 'output_volume' : output_volume}
    if output_volume is None:
        raise ValueError('output_volume must not be None. Please ensure the output volume is supplied in the config dict')
    if unet is None:
        raise ValueError('unet must not be none. Please ensure a unet was loaded and added to the config dict in the config function')
    process_chunks(input_volume, chunk_size, output_volume, margin, 
                   predict_chunk_feature_map, config=config)
    ws.segment_output_image(output_volume, affinities_channels=(0, 1, 2), 
                            thresholding_channel=3, centroids_channel=4,
                            out=current_output.ravel())
    output_volume[:] = 0


# ----------
# U-net Mask
# ----------

def unet_mask(
        napari_viewer, 
        input_volume_layer: napari.layers.Image,
        save_dir: Union[str, None] = None,
        name: str = 'labels-prediction',
        unet_or_config_file: Union[str, None] = None,
        layer_reference: Union[str, None] = None,
        chunk_size: tuple = (10, 256, 256),
        margin: tuple = (1, 64, 64),
    ):
    '''
    ...

    Parameters
    ----------
    napari_viewer: napari.Viewer
        The viewer into which the resultant segmentation will be placed.
    input_volume_layer: napari.layers.Image
        The viewer layer with the image to be processed.
    save_dir: str or None (None)
        The directory into which to save the outputted segmentation. 
    name: str ('labels-prediction')
        The name with which to save the data. Also the name of the 
        resultant segmentation layer that will be displayed in the 
        viewer. 
    config_file: str or None (None)
        Path to a json file with any additional parameters. I'm sure
        There is  a more elegant way to do this through the GUI, 
        but I don't have the time to work on this for now. 
    layer_reference: str or None (None)
        If you need information from one of the layers, this is the
        layer to get it from. This is fed into the custom config function. 
        This function can get a unet from layer metadata. 
    chunk_size: tuple (10, 256, 256)
         This tells you how big the chunks to be processed are
    margin: tuple (1, 64, 64)
        This tells you how much overlap between chunks is used
    '''
    segmentation_wrapper(unet_mask_for_chunks, unet_mask_prep_config, 
                         napari_viewer, input_volume_layer, save_dir, name, 
                         unet_or_config_file, layer_reference, chunk_size, margin)


# UM - Processing function
# ------------------------

def unet_mask_for_chunks(
        input_volume, 
        current_output, 
        chunk_size, 
        margin, 
        # specific
        output_volume,
        unet,
        **kwargs
        ):
    '''
    Function that inserts a unet-derived mask into current output.
    
    Parameters
    ----------
    input_volume: array 
        This is the image to process
    output_volume: array
        This is the array that is used for intermediate processing
        steps (e.g., unet-derived feature map, LoG processed image).
        In this case it is a feature map showing the probability that
        a pixel belongs to a segmentable object.
    chunk_size: tuple
        This tells you how big the chunks to be processed are
    margin: tuple
        This tells you how much overlap between chunks is used
    current_output: array
        This is the 3D array that stores the labels for the particular
        slice. The labels are then written into the labels layer data, 
        which is then displayed on the viewer/ 
    unet: str
        Path to the unet to load to process the data. This will come
        from the config file. 
    use_default_unet: bool
        Do you want to get a mask from the default unet (which has been
        trained to find platelets). There is a mask in channel 3. 
        This may or may not be useful. Regardless, it is accessible 
        through the unet-mask.  
    '''
    default_only_mask = unet == 'default' or unet is None 
    config = {
        'unet' : unet, 
        'default_only_mask' : default_only_mask
    }
    process_chunks(input_volume, chunk_size, output_volume, margin, 
                   predict_chunk_feature_map, 
                   prep_kwarg_function=load_unet, config=config)
    mask = ws._get_mask(output_volume)
    current_output[:, ...] = mask


# UM - Config
# -----------

def unet_mask_prep_config(
        unet, 
        which_unet, 
        reference_layer, 
        config, 
    ):
    if which_unet == 'default' or which_unet is None:
        unet = None
    if which_unet == 'file':
        unet = unet
    elif which_unet == 'labels layer':
        unet = reference_layer.metadata['unet']




# ---------
# Otsu Mask 
# ---------

def otsu_mask(
        napari_viewer, 
        input_volume_layer: napari.layers.Image,
        save_dir: Union[str, None] = None,
        name: str = 'labels-prediction',
        config_file: Union[str, None] = None,
        layer_reference: Union[str, None] = None,
        chunk_size: tuple = (10, 256, 256),
        margin: tuple = (1, 64, 64),
    ):

    '''
    Funtion for finding a simple mask using Otsu thresholding

    Parameters
    ----------
    napari_viewer: napari.Viewer
        The viewer into which the resultant segmentation will be placed.
    input_volume_layer: napari.layers.Image
        The viewer layer with the image to be processed.
    save_dir: str or None (None)
        The directory into which to save the outputted segmentation. 
    name: str ('labels-prediction')
        The name with which to save the data. Also the name of the 
        resultant segmentation layer that will be displayed in the 
        viewer. 
    config_file: str or None (None)
        Path to a json file with any additional parameters. I'm sure
        There is  a more elegant way to do this through the GUI, 
        but I don't have the time to work on this for now. 
    layer_reference: str or None (None)
        If you need information from one of the layers, this is the
        layer to get it from. This is fed into the custom config function. 
        This function doesn't use this. 
    chunk_size: tuple (10, 256, 256)
         This tells you how big the chunks to be processed are
    margin: tuple (1, 64, 64)
        This tells you how much overlap between chunks is used
    '''
    segmentation_wrapper(otsu_mask_for_chunks, otsu_mask_prep_config, 
                         napari_viewer, input_volume_layer, save_dir, name, config_file,
                         layer_reference, chunk_size, margin)



def otsu_mask_for_chunks(
        input_volume, 
        current_output, 
        chunk_size, 
        margin,
        # specific
        gaus_sigma,
        **kwargs
        ):
    '''
    Function that inserts a the results of an affinity watershed into 
    current output

    Parameters
    ----------
    input_volume: array 
        This is the image to process
    output_volume: array
        This is the array that is used for intermediate processing
        steps (e.g., unet-derived feature map, LoG processed image). 
        In this case, it is a smoothed image (using a median filter
        because I like to live dangerously). 
    chunk_size: tuple
        This tells you how big the chunks to be processed are
    margin: tuple
        This tells you how much overlap between chunks is used
    current_output: array
        This is the 3D array that stores the labels for the particular
        slice. The labels are then written into the labels layer data, 
        which is then displayed on the viewer. 
    '''
    mask = ws._get_mask(input_volume)
    current_output[:, ...] = mask


def otsu_mask_prep_config(gaus_sigma, **kwargs):
    pass




# --------------
# Blob Watershed
# --------------

def blob_watershed(
        napari_viewer, 
        input_volume_layer: napari.layers.Image,
        save_dir: Union[str, None] = None,
        name: str = 'labels-prediction',
        config_file: Union[str, None] = None,
        layer_reference: Union[str, None] = None,
        chunk_size: tuple = (10, 256, 256),
        margin: tuple = (1, 64, 64),
    ):
    '''
    Funtion for finding a simple mask using Otsu thresholding

    Parameters
    ----------
    napari_viewer: napari.Viewer
        The viewer into which the resultant segmentation will be placed.
    input_volume_layer: napari.layers.Image
        The viewer layer with the image to be processed.
    save_dir: str or None (None)
        The directory into which to save the outputted segmentation. 
    name: str ('labels-prediction')
        The name with which to save the data. Also the name of the 
        resultant segmentation layer that will be displayed in the 
        viewer. 
    config_file: str or None (None)
        Path to a json file with any additional parameters. I'm sure
        There is  a more elegant way to do this through the GUI, 
        but I don't have the time to work on this for now. 
    layer_reference: str or None (None)
        If you need information from one of the layers, this is the
        layer to get it from. This is fed into the custom config function. 
        This function doesn't use this. 
    chunk_size: tuple (10, 256, 256)
         This tells you how big the chunks to be processed are
    margin: tuple (1, 64, 64)
        This tells you how much overlap between chunks is used
    '''
    segmentation_wrapper(blob_watershed_for_chunks, blob_watershed_prep_config, 
                         napari_viewer, input_volume_layer, save_dir, name, config_file,
                         layer_reference, chunk_size, margin)


def blob_watershed_for_chunks(
        input_volume, 
        current_output, 
        chunk_size, 
        margin,
        # specific
        min_sigma,
        max_sigma, 
        num_sigma, 
        threshold, 
        gaus_sigma
        ):
    '''
    Function that inserts a the results of an affinity watershed into 
    current output

    Parameters
    ----------
    input_volume: array 
        This is the image to process
    output_volume: array
        This is the array that is used for intermediate processing
        steps (e.g., unet-derived feature map, LoG processed image). 
        In this case, it is a smoothed image (using a median filter
        because I like to live dangerously). 
    chunk_size: tuple
        This tells you how big the chunks to be processed are
    margin: tuple
        This tells you how much overlap between chunks is used
    current_output: array
        This is the 3D array that stores the labels for the particular
        slice. The labels are then written into the labels layer data, 
        which is then displayed on the viewer. 
    min_sigma: scalar or sequence of scalars
        Minimum sigma of a gaussian kernel to use for the laplacian 
        of gaussian (LoG) blob detection. The smaller this is, the smaller 
        the smallest objects you detect will be.
    max_sigma: scalar or sequence of scalars
        Maximum sigma of a gaussian kernel to use for the LoG blob 
        detection. The larger this is, the larger the smallest 
        objects you detect will be.
    num_sigma: int
        The number of intermediate values of standard deviations 
        to consider between min_sigma and max_sigma for LoG. 
    threshold: float or None
        The absolute lower bound for scale space maxima for LoG. 
        Local maxima smaller than threshold are ignored.
    gaus_sigma: scalar or sequence of scalars
        The sigma to use for the gaussian smoothing used to make
        the mask from the image.
    '''
    # find the blob seeds for the watershed
    markers = blob_log(input_volume, min_sigma=min_sigma, max_sigma=max_sigma, 
                         num_sigma=num_sigma, threshold=threshold)
    
    distance = ndi.distance_transform_edt(input_volume)
    mask = ws._get_mask(input_volume, gaus_sigma)
    labels = watershed(-distance, markers, mask=mask)
    current_output[:, ...] = labels
    

def blob_watershed_prep_config(
        max_sigma, 
        num_sigma, 
        threshold, 
        gaus_sigma, 
        **kwargs
        ):
    if min_sigma is None:
        min_sigma = 1
    if max_sigma is None:
        max_sigma = 30
    if num_sigma is None:
        num_sigma = 10
    if threshold is None:
        threshold = 0.1
    if gaus_sigma is None:
        gaus_sigma = 2
    new_kwargs = {
        'max_sigma' : max_sigma, 
        'min_sigma' : min_sigma,
        'num_sigma' : num_sigma, 
        'threshold' : threshold, 
        'gaus_sigma' : gaus_sigma
    }
    return new_kwargs



# ------------------
# DoG Blob Watershed
# ------------------

def dog_blob_watershed(
        napari_viewer, 
        input_volume_layer: napari.layers.Image,
        save_dir: Union[str, None] = None,
        name: str = 'labels-prediction',
        config_file: Union[str, None] = None,
        layer_reference: Union[str, None] = None,
        chunk_size: tuple = (10, 256, 256),
        margin: tuple = (1, 64, 64),
        debug: bool =False
    ):
    '''
    Funtion for segmenting an image using DoG blob detection

    Parameters
    ----------
    napari_viewer: napari.Viewer
        The viewer into which the resultant segmentation will be placed.
    input_volume_layer: napari.layers.Image
        The viewer layer with the image to be processed.
    save_dir: str or None (None)
        The directory into which to save the outputted segmentation. 
    name: str ('labels-prediction')
        The name with which to save the data. Also the name of the 
        resultant segmentation layer that will be displayed in the 
        viewer. 
    config_file: str or None (None)
        Path to a json file with any additional parameters. I'm sure
        There is  a more elegant way to do this through the GUI, 
        but I don't have the time to work on this for now. 
    layer_reference: str or None (None)
        If you need information from one of the layers, this is the
        layer to get it from. This is fed into the custom config function. 
        This function doesn't use this. 
    chunk_size: tuple (10, 256, 256)
         This tells you how big the chunks to be processed are
    margin: tuple (1, 64, 64)
        This tells you how much overlap between chunks is used
    '''
    segmentation_wrapper(dog_blob_watershed_for_chunks, dog_blob_watershed_prep_config, 
                         napari_viewer, input_volume_layer, save_dir, name, config_file,
                         layer_reference, chunk_size, margin, debug)


def dog_blob_watershed_for_chunks(
        input_volume, 
        current_output, 
        chunk_size, 
        margin,
        # specific
        min_sigma,
        max_sigma, 
        threshold, 
        **kwargs
        ):
    '''
    Function that inserts the results of a DoG blob detection into
    current_output.

    Parameters
    ----------
    input_volume: array 
        This is the image to process
    chunk_size: tuple
        Not used here.
    margin: tuple
        Not used here. 
    current_output: array
        This is the 3D array that stores the labels for the particular
        slice. The labels are then written into the labels layer data, 
        which is then displayed on the viewer. 
    min_sigma: scalar or sequence of scalars
        Minimum sigma of a gaussian kernel to use for the difference 
        of gaussian (DoG) blob detection. The smaller this is, the smaller 
        the smallest objects you detect will be.
    max_sigma: scalar or sequence of scalars
        Maximum sigma of a gaussian kernel to use for the DoG blob 
        detection. The larger this is, the larger the smallest 
        objects you detect will be.
    threshold: float or None
        The absolute lower bound for scale space maxima for LoG. 
        Local maxima smaller than threshold are ignored.
    gaus_sigma: scalar or sequence of scalars
        The sigma to use for the gaussian smoothing used to make
        the mask from the image.
    '''
    # find the blob seeds for the watershed
    input_volume = np.pad(input_volume, pad_width=1)
    dog = dog_image(input_volume, min_sigma, max_sigma)
    mask = dog > threshold
    markers = blob_dog(input_volume, min_sigma=min_sigma, max_sigma=max_sigma, 
                        threshold=threshold)
    distance = ndi.distance_transform_edt(input_volume)
    #local_maxi = peak_local_max(distance, min_distance=3, labels=input_volume)
    centroids = np.zeros(distance.shape, dtype=bool)
    idx = tuple(markers.T.astype(int))[:-1] # there's an extra index
    centroids[idx] = True
    markers, num_objects = ndi.label(centroids) #, structure=np.ones((3,3,3)))
    labels = watershed(-distance, markers, mask=mask)
    #distance = ndi.distance_transform_edt(input_volume)
    #mask = ws._get_mask(input_volume, gaus_sigma)
    #labels = watershed(-distance, markers, mask=mask)
    current_output[:, ...] = labels
    


def dog_blob_watershed_prep_config(
        input_volume_layer, # not used here
        unet_or_config_file, 
        reference_layer, # not used here
        max_sigma=1.5, 
        min_sigma=1,
        threshold=0.02, 
        ):
    if unet_or_config_file is not None:
        config = read_config_json(unet_or_config_file)
        val = config.get['max_sigma']
        max_sigma = val if val is not None else max_sigma
        val = config.get['min_sigma']
        min_sigma = val if val is not None else min_sigma
        val = config.get['threshold']
        threshold = val if val is not None else threshold
    new_kwargs = {
        'max_sigma' : max_sigma, 
        'min_sigma' : min_sigma,
        'threshold' : threshold, 
    }
    return new_kwargs


def dog_image(input_vol, sigma_min, sigma_max):
    dog = gaussian(input_vol, sigma_min) - gaussian(input_vol, sigma_max)
    return dog

# ---------------
# Common funcions
# ---------------


def read_config_json(path_to_json):
    with open(path_to_json, 'r') as f:
        data = json.load(f)
    return data


def output_vol_function_default(data):
    shape = (1, ) * (4 - data.ndim) + data.shape
    output_volume = np.zeros(data.shape, dtype=np.float32) 
    return output_volume



def segmentation_wrapper(
        processing_function: Callable, 
        config_prep_function: Callable,
        napari_viewer, 
        input_volume_layer: napari.layers.Image,
        save_dir: Union[str, None],
        name: str,
        network_or_config_file: Union[str, None],
        layer_reference: Union[str, None],
        chunk_size: tuple,
        margin: tuple,
        debug: bool =False
        ):
    """
    Function that takes a function for processing, preparing the configuration
    dict (key word arguments). 

    Parameters
    ----------
    processing_function: Callable
        The function you want to use to process the data. This function
        should take key word arguments (**kwargs). The function must
        take the following arguments in this order:
            - input_volume: array (3D) - the image frame to be processed
            - current_output: array (3D) - write the labels for the current frame here
            - chunk_size: tuple (3D) - size of chunks  
            - margin: tuple (3D) - size of margin
    config_prep_function: Callable (None)
        If not None, this function will be called before entering the 
        segmentation loop. It prepares a dictionary to be fed to the 
        process functioning as kew word arguments (**kwargs) when it
        is executed. 
            - input_volume_layer: napari.layers.Image - image layer to be processed
            - network_or_config_file: str - where to find the file you want to use
            - layer_reference: napari.layers.Layer - layer with metadata of interest
    napari_viewer: napari.Viewer
        The napari viewer. 
    input_volume_layer: napari.layers.Image
        The imput layer you want to segment. 
    save_dir: str or None
        The directory to which you want to save the segmentation to. 
    name: str
        The name you want to save the segmentation under. The segmentation
        will automatically be saved as a zarr with the extention .zarr. You 
        don't need to include an extension in the name. This name will also
        be used as the layer name. 
    network_or_config_file: str or None
        This is the file containing a neural network of 
    layer_reference: str or None
    chunk_size: tuple
    margin: tuple
    output_vol_function: function or None
    """
    
    viewer = napari_viewer

    # get the config dict
    # -------------------
    config = config_prep_function(input_volume_layer, network_or_config_file, layer_reference)
    if config is None:
        config = {}

    # Prepare data
    # ------------
    shape_4D = np.broadcast_shapes((1,) * 4, input_volume_layer.data.shape)
    data = np.reshape(input_volume_layer.data, shape_4D)
    scale = viewer.layers[0].scale[-3:] # lazy assumption that all layers have the same scale 
    translate = viewer.layers[0].translate[-3:] # and same translate 
    ndim = len(chunk_size)

    # Output labels
    # --------------
    # get the output labels and layer for output segmentation
    output_labels = zarr.zeros(
        shape=shape_4D, 
        chunks=(1,) + data.shape[-3:], 
        dtype=np.int32, 
        )
    output_layer = viewer.add_labels(
            output_labels,
            name=name,
            scale=scale,
            translate=translate,
            )

    max_t = data.shape[0] - 1

    # for yeilds
    # ----------
    def handle_yields(yielded_val):
        viewer.dims.current_step = (yielded_val, 0, 0, 0)
        print(f"Segmented t = {yielded_val}")
        if yielded_val == max_t and save_dir is not None:
            save_path = os.path.join(str(save_dir), name + '.zarr')
            zarr.save(save_path, output_labels)


    # for errors
    # ----------
    def handle_errors(errored_val):
        raise errored_val
    

    # define the worker
    # -----------------
    launch_worker = thread_worker(
        segmentation_loop,
        progress={'total': data.shape[0], 'desc': 'thread-progress'},
        # this does not preclude us from connecting other functions to any of the
        # worker signals (including `yielded`)
        connect={'yielded': handle_yields, 'errored': handle_errors},
    )

    # Launch the worker
    # -----------------
    if not debug:
        worker = launch_worker( # this is where the process dies... "There appear to be 2 leaked semaphore objects ..."
            viewer, 
            data, 
            chunk_size, 
            margin, 
            ndim, 
            output_labels, 
            processing_function,
            config
        )
    else:
        for t in segmentation_loop(viewer, data, chunk_size, margin, ndim, output_labels, processing_function, config):
            print(f'Segmented frame {t}')

    # Save the data
    # -------------
    if save_dir is not None and not debug:
        save_path = os.path.join(str(save_dir), name + '.zarr')
        zarr.save(save_path, output_labels)

    return output_layer


def segmentation_loop(
        viewer,
        data, 
        chunk_size, 
        margin, 
        ndim, 
        output_labels, 
        processing_function,
        config
    ):
    '''
    This function runs the segmentation function on one 3D frame at a time. 
    The function yeilds the current time point when done and this is printed
    in the main thread. 

    Parameters
    ----------
    data: np.array or zarr.array
    viewer: napari.Viewer
    output_volume: np.array or zarr.array
    chunk_size: tuple of len ndim
    margin: tuple of len ndim
    ndim: int
    output_labels: np.array or zarr.array
    processing_function: func
        Function for producing the full segmentation
    config: dict
        Key word arguments for processing function
    '''
    for t in range(data.shape[0]):
        #print(t) 
        slicing = (t, slice(None), slice(None), slice(None))
        input_volume = np.asarray(data[slicing]).astype(np.float32)
        if input_volume.min() == 0:
            input_volume = remove_sum_zero_slices(input_volume)
            #random_vol = np.random.normal(input_volume.mean(), size=input_volume.shape)
            #input_volume = np.where(input_volume == 0, random_vol, input_volume).astype(np.float32)
        input_volume /= np.max(input_volume)
        current_output = np.pad(
            np.zeros(data.shape[1:], dtype=np.uint32), # not sure if this will work
            1,
            mode='constant',
            constant_values=0,
            )
        crop = tuple([slice(1, -1),] * ndim)  # yapf: disable
        # predict using unet
        processing_function(input_volume, current_output, chunk_size, margin, **config)
        #current_output[crop] = np.where(input_volume == 0, 0, current_output[crop])
        output_labels[t, ...] = current_output[crop]
        yield t


def remove_sum_zero_slices(input_volume):
    for ax_i in range(input_volume.ndim):
            ax_nonzero_rows = []
            for i in range(input_volume.shape[ax_i]):
                s = [slice(None) for i in range(input_volume.ndim)]
                s[ax_i] = slice(i, i+1)
                s = tuple(s)
                if input_volume[s].sum() != 0:
                    ax_nonzero_rows.append(i)
            s = [slice(None) for i in range(input_volume.ndim)]
            s[ax_i] = ax_nonzero_rows
            s = tuple(s)
            input_volume = input_volume[s]
    return input_volume


# ------------
# summary dict 
# ------------

# to be used in segment_data doc widget
segmenters = {
    'affinity-unet-watershed' : affinity_unet_watershed, 
  #  'unet-mask' : unet_mask, 
   # 'otsu-mask' : otsu_mask,
  #  'LoG-blob-watershed' : blob_watershed,
    'DoG-blob-watershed' : dog_blob_watershed
}




# coding: utf-8
import os
import itertools
import torch
from tqdm import tqdm
import numpy as np
import toolz as tz
import napari
from napari.qt import thread_worker
from skimage.exposure import rescale_intensity
import platform

from . import unet as unet_mod
from . import watershed as ws

# ----------------------------------------------------------------------------
# The following was copied from https://github.com/jni/platelet-unet-watershed
# ----------------------------------------------------------------------------
IGNORE_CUDA = False

DEFAULT_UNET_PATH = os.path.join(
            os.path.dirname(__file__), 'data/211309_151717_unet_z-1_y-1_x-1_m_c.pt'
            )

def load_unet(u_state_fn=DEFAULT_UNET_PATH):
    if u_state_fn is None:
        u_state_fn = DEFAULT_UNET_PATH
    u = unet_mod.UNet(in_channels=1, out_channels=5)
    device = get_device()
    map_location = device  # for loading the pre-existing unet
    if torch.cuda.is_available() and not IGNORE_CUDA:
        u.cuda()
        map_location = None
    u.load_state_dict(torch.load(u_state_fn, map_location=map_location))
    return u


def make_chunks(arr_shape, chunk_shape, margin):
    ndim = len(arr_shape)
    if type(margin) == int:
        margin = [margin] * ndim
    starts = []
    crops = []
    for dim in range(ndim):
        arr = arr_shape[dim]
        chk = chunk_shape[dim]
        mrg = margin[dim]
        start = np.arange(0, arr - 2*mrg, chk - 2*mrg)
        start[-1] = arr - chk
        if len(start) > 1 and start[-1] == start[-2]:
            # remove duplicates in case last step is perfect
            start = start[:-1]
        starts.append(start)
        crop = np.array([(mrg, chk - mrg),] * len(start))  # yapf: disable
        crop[0, 0] = 0
        crop[-1, 0] = chk - (arr - np.sum(crop[:-1, 1] - crop[:-1, 0]))
        crop[-1, 1] = chk
        crops.append(crop)
    chunk_starts = list(itertools.product(*starts))
    chunk_crops = list(itertools.product(*crops))
    return chunk_starts, chunk_crops


def process_chunks(
        input_volume,
        chunk_size,
        output_volume,
        margin,
        process_data_function,
        config=None
        ):
    # kewy word arguments for the processing function
    if config is None:
        config = {}
    # get the chunks
    ndim = len(chunk_size)
    chunk_starts, chunk_crops = make_chunks(
            input_volume.shape[-ndim:], chunk_size, margin=margin
            )
    # loop to go through the chunks - make parallel if get the chance 
    for start, crop in tqdm(list(zip(chunk_starts, chunk_crops))):
        sl = tuple(
                slice(start0, start0 + step)
                for start0, step in zip(start, chunk_size)
                )
        sl = (slice(None), ) + sl
        predicted_array = process_data_function(input_volume, sl, **config)
        # check the ndims to adjust crop
        p_dim = predicted_array.ndim
        o_dim = output_volume.ndim
        cr = (slice(None),) * (p_dim - o_dim) + tuple(slice(i, j) for i, j in crop), 
        cr = cr[0]
        # get the desired crop
        pred_c = (0, ) + cr
        output_volume[sl][cr] = predicted_array[pred_c]
    return output_volume



def predict_chunk_feature_map(
        input_volume, 
        sl, 
        unet=False, 
        default_only_mask=False, 
        **kwargs
        ):
    '''
    Parameters
    ----------
    input_volume: array
        Image to be processed by the unet.
    u: unet.UNet
        The U Net you will produce the feature map with. 
    sl: tuple of slice 
        Slice to take from the input_volume. Needs to have the 
        same ndim as input_volume.
    '''
    assert unet != False, 'Please ensure a unet is loaded and supplied'
    sl = sl[1:]
    tensor = torch.from_numpy(input_volume[sl][np.newaxis, np.newaxis])
    if torch.cuda.is_available() and not IGNORE_CUDA:
        tensor = tensor.cuda()
    predicted_array = unet(tensor).detach().cpu().numpy() #
    if default_only_mask: # if using default network but only the mask channel is wanted
        predicted_array = predicted_array[3, ...]
    return predicted_array



def get_device():
    if torch.backends.mps.is_available() and torch.backends.mps.is_built():
        device = torch.device("mps")
    else:
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    return device



# ----------------
# To Be Deprecated
# ----------------


def predict_output_chunks(
        unet, # now a string: 'path/to/unet' - only used if use_default_unet == False
        input_volume,
        chunk_size,
        output_volume,
        margin=0,
        use_default_unet=True, 
        mask_unet=False
        ):
    if unet is None:
        u = load_unet()
    else:
        u = load_unet(unet)
    #print(u)
    ndim = len(chunk_size)
    chunk_starts, chunk_crops = make_chunks(
            input_volume.shape[-ndim:], chunk_size, margin=margin
            )
    # add a for-loop for higher dims (input-volume.shape[:-ndim])
    # OR, do it one level up
    for start, crop in tqdm(list(zip(chunk_starts, chunk_crops))):
        sl = tuple(
                slice(start0, start0 + step)
                for start0, step in zip(start, chunk_size)
                )
        tensor = torch.from_numpy(input_volume[sl][np.newaxis, np.newaxis])
        if torch.cuda.is_available() and not IGNORE_CUDA:
            tensor = tensor.cuda()
        predicted_array = u(tensor).detach().cpu().numpy()
        #print(predicted_array.shape)
        # add slice(None) for the 5 channels
        cr = (slice(None),) + tuple(slice(i, j) for i, j in crop), 
        if not mask_unet:
            output_volume[(slice(None),) + sl][cr] = predicted_array[(0,) + cr]
        elif mask_unet and use_default_unet:
            # the mask is in channel 5
            output_volume[(slice(3,4), ) + sl][cr] = predicted_array[(0,) + cr]
        else:
            # assume that the output volume has only one channel
            output_volume[sl][cr] = predicted_array[(0,) + cr]
        #yield
    #print('op: ', output_volume.shape, np.max(output_volume))
    return output_volume




from .custom_loss import WeightedBCELoss, DiceLoss, EpochwiseWeightedBCELoss, \
    channel_losses_to_dict, ChannelwiseLoss
from datetime import datetime
import dask.array as da
from .helpers import LINE, write_log
import napari
import numpy as np
import os
import pandas as pd
from .plots import save_loss_plot, save_channel_loss_plot
from tifffile import TiffWriter 
import torch 
import torch.nn as nn
import torch.optim as optim
from .train_io import get_train_data, load_train_data, load_dask_as_tensor, load_tensor_from_zarr
from tqdm import tqdm
from .unet import UNet, ForkedUNet


# DICE seems to be more common but BCE Loss is also used for 
#   image segmentation (Jadon et al., 2020, arXiv) and seems 
#   to give better results here


def train_unet(
               # training data
               x,
               vx,  
               y, 
               vy,
               ids=None,
               vids=None, 
               # output information
               out_dir=None, 
               name='my-unet', 
               channels=None,
               # training variables
               validate=True,
               log=True,
               epochs=3, 
               lr=0.01, 
               loss_function='BCELoss', 
               chan_weights=None, # for weighted BCE
               weights=None,
               update_every=20, 
               losses=None, 
               chan_losses=None,
               # network architechture
               fork_channels=None,
               chan_final_activations=None,
               **kwargs
               ):
    '''
    Train a basic U-Net on affinities data.

    Parameters
    ----------
    xs: list of torch.tensor
        Input images for which the network will be trained to predict
        inputted labels.
    ys: list of torch.tensor
        Input labels that represent target output that the network will
        be trained to predict.
    ids: list of str
        ID strings corresponding to xs and ys (as they are named on disk). 
        Used for saving output.
    out_dir: str
        Directory to which to save network output
    suffix: str
        Suffix used in naming pytorch state dictionary file
    channels: tuple of str or None
        Names of output channels to be used for labeling channelwise
        loss columns in output loss csv. If none, names are generated.
    v_xs: list of torch.Tensor or None
        Validation images
    v_y: list of torch.Tensor or None
        Validation labels
    v_ids: list of str or None
        Validation IDs
    validate: bool
        Will a validation be done at the end of every epoch?
    log: bool
        Will a log.txt file containing all console print outs be saved?
    epochs: int
        How many times should we go through the training data?
    lr: float
        Learning rate for Adam optimiser
    loss_function: str
        Which loss function will be used for training & validation?
        Current options include:
            'BCELoss': Binary cross entropy loss
            'WeightedBCE': Binary cross entropy loss whereby channels are weighted
                according to chan_weights parameter. Quick way to force network to
                favour learning information about a given channel/s.
            'DiceLoss': 1 - DICE coefficient of the output-target pair
    chan_weights: tuple of float
        WEIGHTEDBCE: Weights for BCE loss for each output channel. 
    weights: None or nn.Model().state_dict()
        Prior weights with which to initalise the network.
    update_every: int
        Determines how many batches are processed before printing loss

    Returns
    -------
    unet: UNet (unet.py)

    Notes
    -----
    When data is loaded from a directory, it will be recognised according
    to the following naming convention:

        IDs: YYMMDD_HHMMSS_{digit/s} 
        Images: YYMMDD_HHMMSS_{digit/s}_image.tif
        Affinities: YYMMDD_HHMMSS_{digit/s}_labels.tif
    
    E.g., 210309_152717_7_image.tif, 210309_152717_7_labels.tif

    For each ID, a labels and an image file must be found or else an
    assertion error will be raised.
    '''
    # Save?
    save_output = out_dir is not None
    print('Output will be saved: ', save_output)
    print('Save directory: ', out_dir)
    # Unique identifiers
    if ids is None:
        ids = [name + f'_{i}' for i in range(len(x))]
    if vids is None:
        vids = [name + f'_val_{i}' for i in range(len(vx))]
    # Device
    device_name = 'cuda' if torch.cuda.is_available() else 'cpu'
    device = torch.device(device_name)
    # initialise U-net
    if fork_channels is None:
        unet = UNet(out_channels=len(channels), 
                    chan_final_activations=chan_final_activations).to(device, 
                                                                      dtype=torch.float32)
    else:
        unet = UNet(out_channels=fork_channels, 
                    chan_final_activations=chan_final_activations).to(device, 
                                                                      dtype=torch.float32)
    # load weights if applicable 
    weights_are = _load_weights(weights, unet)
    # define the optimiser
    optimiser = optim.Adam(unet.parameters(), lr=lr)
    # define the loss function
    loss = _get_loss_function(loss_function, chan_weights, 
                              device, losses, chan_losses)
    # get the dictionary that will be converted to a csv of losses
    #   contains columns for each channel, as we record channel-wise
    #   BCE loss in addition to the loss used for backprop
    channels = _index_channels_if_none(channels, x) 
    loss_dict = _get_loss_dict(channels)
    v_loss = _get_loss_function(loss_function, chan_weights, 
                                device, losses, chan_losses)
    validation_dict = {
        'epoch' : [], 
        'validation_loss' : [], 
        'data_id' : [], 
        'batch_id': []
    }
    if validate:
        no_iter = (epochs * len(x)) + ((epochs + 1) * len(vx))
    else:
        no_iter = epochs * len(x)
    # print the training into and log if applicable 
    bce_weights = _bce_weights(loss) # gets weights if using WeightedBCE
    _print_train_info(loss_function, bce_weights, epochs, lr, 
                      weights_are, device_name, out_dir, log, 
                      chan_losses, losses, channels, fork_channels)
    # loop over training data 
    v_y_hats = _train_loop(no_iter, epochs, x, y, ids, device, unet, 
                           out_dir, optimiser, loss, loss_dict,  
                           validate, vx, vy, vids, validation_dict, 
                           v_loss, update_every, log, name, channels, 
                           save_output)
    if save_output:
        print('Saving Final Results...')
        unet_path = _save_final_results(unet, out_dir, name, ids, validate, 
                                        loss_dict, v_y_hats, vids, 
                                        validation_dict)
    #_plots(out_dir, name, loss_function, validate) # 2 leaked semaphore objects... pytorch x mpl??
    return unet, unet_path



def _plots(out_dir, name, loss_function, validate):
    l_path = os.path.join(out_dir, 'loss_' + name + '.csv')
    v_path = None
    if validate:
        vl_path = os.path.join(out_dir, 'validation-loss_' + name + '.csv')
    save_loss_plot(l_path, loss_function, v_path=vl_path, show=False)
    save_channel_loss_plot(l_path, show=False)



def _get_loss_function(loss_function, chan_weights, device, 
                       losses, chan_losses):
    # define the loss function
    if loss_function == 'BCELoss':
        loss = nn.BCELoss()
    elif loss_function == 'DiceLoss':
        loss = DiceLoss()
    elif loss_function == 'WeightedBCE':
        loss = WeightedBCELoss(chan_weights=chan_weights, device=device)
    #elif loss_function == 'BCECentrenessPenalty':
       # loss = BCELossWithCentrenessPenalty()
    elif loss_function == 'EpochWeightedBCE':
        loss = EpochwiseWeightedBCELoss(weights_list=chan_weights, device=device)
    elif loss_function == 'Channelwise':
        loss = ChannelwiseLoss(losses, chan_losses, device)
    elif loss_function == 'MSELoss':
        loss = nn.MSELoss()
    else:
        m = 'Valid loss options are BCELoss, WeightedBCE, and DiceLoss'
        raise ValueError(m)
    return loss


def _load_weights(weights, unet):
    weights_are = 'naive'
    if weights is not None:
        unet.load_state_dict(weights)
        weights_are = 'pretrained'
    return weights_are


def _index_channels_if_none(channels, x):
    if channels is None:
        new_chans = ['channel_' + str(i) for i in range(x[0].shape[1])]
        return tuple(new_chans)
    else:
        return channels


def _get_loss_dict(channels):
    loss_dict = {'epoch' : [], 
                 'batch_num' : [], 
                 'loss' : [], 
                 'data_id' : []}
    for c in channels:
        loss_dict[c] = []
    return loss_dict


def _bce_weights(loss):
    bce_weights = None
    try:
        bce_weights = loss.chan_weights.data
    except:
        pass
    return bce_weights


def _print_train_info(loss_function, bce_weights, epochs, lr, 
                     weights_are, device_name, out_dir, log, 
                     chan_losses, losses, channels, fork_channels):
    s = LINE + '\n' + f'Loss function: {loss_function} \n'
    if bce_weights is not None:
        s = s + f'    Loss function channel weights: {bce_weights} \n'
    if losses is not None:
        chan_bd = []
        for c in chan_losses:
            if isinstance(c, slice):
                chan_bd.append(f'[{c.start}, {c.stop})')
            else:
                chan_bd.append(str(c))
        for i, l in enumerate(losses):
            s = s + f'    Loss for channels {chan_bd[i]}: {l}\n'
    s = s + 'Optimiser: Adam \n' + f'Learning rate: {lr} \n'
    s = s + LINE + '\n' 
    s = s + f'Training {weights_are} U-net for {epochs} '
    s = s + 'epochs with batch size 1 \n'
    s = s + f'Device: {device_name} \n'
    if channels is not None:
        s = s + f'Channels: {channels}\n'
    if fork_channels is not None:
        s = s + f'Channels per fork (according to channel order): {fork_channels}\n' 
    s = s + LINE
    print(s)
    if log:
        write_log(LINE, out_dir)
        write_log(s, out_dir)



def _train_loop(no_iter, epochs, x, y, ids, device, unet, out_dir,
                optimiser, loss, loss_dict,  validate, vx, vy, 
                vids, validation_dict, v_loss, update_every, log, 
                name, channels, save_output):
    v_y_hats = None
    # loop over training data 
    print_mem_every = 3 * update_every
    unet = unet.to(device=device, dtype=torch.float32)
    with tqdm(total=no_iter, desc='unet training') as progress:
        for e in range(epochs):
            _set_epoch_if_epoch_weighted(loss, e)
            if validate and e == 0:
                _set_epoch_if_epoch_weighted(v_loss, e, verbose=False)
                if e == 0:
                    # first validation at the start of the first epoch
                    v_y_hats = _validate(vx, vy, vids, 
                                         device, unet, v_loss, 
                                         progress, log, out_dir, 
                                         validation_dict, e, 0)
            running_loss = 0.0
            for i in range(len(x)):
                l = _train_step(i, x, y, ids, device, unet, optimiser, 
                                loss, loss_dict, e, channels)
                optimiser.step()
                running_loss += l.item()
                progress.update(1)
                if i % update_every == (update_every - 1):
                    s = f'Epoch {e} - running loss: ' 
                    s = s + f'{running_loss / update_every}'
                    print(s)
                    if log:
                        write_log(s, out_dir)
                    running_loss = 0.0
                #if i % print_mem_every == (print_mem_every - 1):
                    #print(torch.cuda.memory_summary())
            if validate:
                # validation at the end of the epoch
                batch_no = ((e + 1) * len(x))
                v_y_hats = _validate(vx, vy, vids, 
                                     device, unet, v_loss, 
                                     progress, log, out_dir, 
                                     validation_dict, e, batch_no)
            if save_output:
                print('Saving Training Checkpoint...')
                _save_checkpoint(unet.state_dict(), out_dir, 
                             f'{name}_epoch-{e}')  
            #torch.cuda.empty_cache()
    return v_y_hats


def _set_epoch_if_epoch_weighted(loss, e, verbose=True):
    if isinstance(loss, EpochwiseWeightedBCELoss):
        loss.current_epoch = e
        if verbose:
            print(f'Channel weights set to {loss.current_weights.data} for epoch {e} ')


def _train_step(i, xs, ys, ids, device, unet, optimiser, 
                loss, loss_dict, e, channels):
    x = load_tensor_from_zarr(i, xs)
    y = load_tensor_from_zarr(i, ys)
    x, y = _prep_x_y(x, y, device)
    optimiser.zero_grad()
    y_hat = unet(x.float())
    l = loss(y_hat, y)
    l.backward()
    optimiser.step()
    loss_dict['epoch'].append(e) 
    loss_dict['batch_num'].append(i)
    loss_dict['loss'].append(l.item())
    loss_dict['data_id'].append(ids[i])
    channel_losses_to_dict(y_hat, y, channels, loss, loss_dict)
    #y_hat = y_hat.detach().cpu().numpy()
    #y_hats.append(y_hat)
    #del y_hat
    del x, y, y_hat
    return l


def _validate(v_xs, v_ys, v_ids, device, unet, v_loss, progress, 
             log, out_dir, validation_dict, e, batch_no):
    validation_loss = 0.0
    with torch.no_grad():
        v_y_hats = []
        for i in range(len(v_xs)):
            v_x = load_tensor_from_zarr(i, v_xs)
            v_y = load_tensor_from_zarr(i, v_ys)
            v_x, v_y = _prep_x_y(v_x, v_y, device)
            v_y_hat = unet(v_x.float())
            v_y_hats.append(v_y_hat)
            vl = v_loss(v_y_hat, v_y)
            validation_loss += vl.item()
            validation_dict['epoch'].append(e)
            validation_dict['validation_loss'].append(vl.item())
            validation_dict['data_id'].append(v_ids[i])
            validation_dict['batch_id'].append(batch_no)
            progress.update(1)
        score = validation_loss / len(v_xs)
        s = f'Epoch {e} - validation loss: {score}'
        print(s)
        if log:
            write_log(s, out_dir)
    return v_y_hats


def _prep_x_y(x, y, device):
    x, y = torch.unsqueeze(x, 0), torch.unsqueeze(y, 0)
    x = torch.unsqueeze(x, 0)
    x, y = x.to(device), y.to(device)
    y = y.type(torch.float32)
    return x, y


def _save_final_results(unet, out_dir, name, ids, validate,
                        loss_dict, v_y_hats, v_ids, validation_dict):
    unet_path = _save_checkpoint(unet.state_dict(), out_dir, name, r=True)
    #_save_output(y_hats, ids, out_dir)
    loss_df = pd.DataFrame(loss_dict)
    loss_df.to_csv(os.path.join(out_dir, 'loss_' + name + '.csv'))
    if validate:
        _save_output(v_y_hats, v_ids, out_dir, name='_validation')
        v_loss_df = pd.DataFrame(validation_dict)
        v_loss_df.to_csv(os.path.join(out_dir, 
                         'validation-loss_' + name + '.csv'))
    return unet_path


def _save_checkpoint(checkpoint, out_dir, name, r=False):
    now = datetime.now()
    d = now.strftime("%y%d%m_%H%M%S")
    name = d + '_unet_' + name + '.pt'
    path = os.path.join(out_dir, name)
    os.makedirs(out_dir, exist_ok=True)
    torch.save(checkpoint, path)
    if r:
        return path


def _save_output(y_hats, ids, out_dir, name=''):
    assert len(y_hats) == len(ids)
    os.makedirs(out_dir, exist_ok=True)
    for i in range(len(y_hats)):
        n = ids[i] + name +'_output.tif'
        p = os.path.join(out_dir, n)
        with TiffWriter(p) as tiff:
            tiff.write(y_hats[i].detach().cpu().numpy())



from .augment import augment_images
import dask.array as da
from datetime import datetime
from .helpers import get_files, log_dir_or_None, write_log, LINE
from .labels import get_training_labels, print_labels_info
import numpy as np
import os
import pandas as pd
from pathlib import Path
import re
from tifffile import TiffWriter, imread
import torch 
from tqdm import tqdm
import zarr
import dask.array as da
from napari.types import ImageData, LabelsData


# -------------------
# Generate Train Data
# -------------------
def get_train_data(
    image_list, 
    gt_list,
    out_dir=None, 
    name='train-unet',
    shape=(10, 256, 256), 
    n_each=100, 
    channels=('z-1', 'y-1', 'x-1', 'centreness'),
    scale=(4, 1, 1), 
    log=True, 
    validation_prop=0.2,
):
    """
    Generate training data from whole ground truth volumes.

    Parameters
    ----------
    image_paths: str
        path to the image zarr file, which must be in tczyx dim format (ome bioformats)
    gt_paths: str
        path to the ground truth zarr file, which must be in tczyx dim format
    shape: tuple of int
        Shape of the test data to 
    channels: tuple of str
        tuple of channels to be added to the training data. 
            Affinities: 'axis-n' (pattern: r'[xyz]-\d+' e.g., 'z-1')
            Centreness: 'centreness'
    scale: tuple of numeric
        Scale of channels. This is used in calculating centreness score.
    log: bool
        Should print out be recorded in out_dir/log.txt?

    Returns
    -------
    chunk_dict: dict
        Dict containing generated training data of the form
        {
            'channels' : {
                'name' : (<channel name>, ...), 
                ...
            },
            'save_dir' : <path/to/random/chunks>,
            'labels_dirs' : {
                'name' : <path/to/random/chunks/labels>, 
                ...
            },
            'ids' : [<chunk id string>, ...]
            'x' : [<image chunk tensor>, ...], 
            'ground_truth' : [<gt chunk tensor>, ...], 
            'ys' : {
                'name'
            }
        }

    Notes
    -----
    if there is only one set of training labels generated, the name
    of the labels (anywhere you see 'name' above) will be 'y'.
    """
    assert len(image_list) == len(gt_list)
    now = datetime.now()
    d = now.strftime("%y%m%d_%H%M%S") + '_' + name
    out_dir = os.path.join(out_dir, d)
    os.makedirs(out_dir, exist_ok=True)
    chunk_dicts = []
    if not isinstance(scale, list):
        scale = [scale, ] * len(image_list)
    for i in range(len(image_list)):
        chunk_dict = get_random_chunks(
            image_list[i], 
            gt_list[i], 
            out_dir, 
            name=name,
            shape=shape, 
            n=n_each, 
            channels=channels,
            scale=scale[i], 
            log=log, 
            image_no=i
        )
        chunk_dicts.append(chunk_dict)
    chunk_dicts = concat_chunk_dicts(chunk_dicts)
    train_dicts = chunk_dict_to_train_dict(chunk_dicts, validation_prop)
    return train_dicts



def get_random_chunks(
                      image_src, 
                      gt_src, 
                      out_dir, 
                      name='unet-training',
                      shape=(10, 256, 256), 
                      n=25, 
                      min_brightness_prop=0.005, 
                      channels=('z-1', 'y-1', 'x-1', 'centreness'),
                      scale=(4, 1, 1), 
                      log=True, 
                      image_no=0
                      ):
    '''
    Obtain random chunks of data from whole ground truth volumes.

    Parameters
    ----------
    image: array like
        Image from which to generate training chunks
    ground_truth: array like
        ground truth segmentation (same shape as image)
    shape: tuple of int
        shape of chunks to obtain
    n: int
        number of random chunks to obtain
    min_brightness_prop:
        minimum cut off sum of affinities for an image.
        As affinities as belong to {0, 1}, this param
        is the number of voxels that boarder labels.
    
    Returns
    -------
    chunk_dict:
 
    '''
    save_output = out_dir is not None
    now = datetime.now()
    d = now.strftime("%y%m%d_%H%M%S") + '_' + name
    if isinstance(image_src, str):
        image = zarr.open_array(image_src)
        im_name = image_src
    else:
        try:
            im_shape = image_src.shape
        except:
            m = f'input image should be a path (str), napari image layer, or array like not {type(image_src)}'
            raise TypeError(m)
        image = image_src
        im_name = f'image_shape-{im_shape}_prepared-{d}'
    image = normalise_data(np.array(image))
    if isinstance(gt_src, str):
        ground_truth = zarr.open_array(gt_src)
        gt_name = gt_src
    else:
        try:
            gt_shape = gt_src.shape
        except:
            m = f'input image should be a path (str), napari image layer, or array like not {type(gt_src)}'
            raise TypeError(m)
        gt_name = f'labels_shape-{gt_shape}_prepared-{d}'
        ground_truth = gt_src
    ground_truth = np.array(ground_truth)
    print(LINE)
    s = f'Generating training data from image: {im_name}, Ground truth: {gt_name}'
    print(s)
    # in the following code, image chunks are np.ndarray
    print('Generating random image chunks...')
    chunk_dict = get_image_chunks(
        image, 
        shape=shape, 
        n=n, 
        min_brightness_prop=min_brightness_prop, 
        image_no=image_no
    )
    chunk_dict['df']['image_no'] = [image_no, ] * len(chunk_dict['df'])
    chunk_dict['df']['image_file'] = [Path(im_name).stem, ] * len(chunk_dict['df'])
    print('Generating training labels...')
    chunk_dict = get_labels_chunks(chunk_dict, ground_truth, 
                                   channels=channels, scale=scale)
    print('Augmenting data...')
    chunk_dict = augment_chunks(chunk_dict)
    # save the chunks whilst np.ndarray
    save_dir = None
    if save_output:
        print('Saving for posterity...')
        save_dir = save_from_chunk_dict(chunk_dict, out_dir, name)
    # convert to tensor 
    #print('Converting to Torch Tensor...')
    #convert_chunks_to_tensor(chunk_dict)
        if log:
            write_log(LINE, save_dir)
            write_log(s, save_dir)
        print(LINE)
        s = f'Obtained {n} {shape} chunks of training data'
        print(s)
        if log:
            write_log(LINE, save_dir)
            write_log(s, save_dir)
    log_dir = log_dir_or_None(log, save_dir)
    print_labels_info(channels, out_dir=log_dir)
    df_path = os.path.join(save_dir, 'start_coords.csv')
    if os.path.exists(df_path):
        df = pd.read_csv(df_path)
        df = pd.concat([df, chunk_dict['df']])
    else:
        df = chunk_dict['df']
    df.to_csv(df_path)
    return chunk_dict


# -----------------------
# Get Random Image Chunks 
# -----------------------

def get_image_chunks(
    image, 
    shape=(10, 256, 256), 
    n=25, 
    min_brightness_prop=0.3, 
    image_no=0,
):
    im = np.array(image)
    assert len(im.shape) == len(shape)
    xs = []
    ids = []
    i = 0
    df = {'z_start' : [],
          'y_start' : [],
          'x_start' : []}
    slices = []
    while i < n:
        dim_randints = []
        for j, dim in enumerate(shape):
            max_ = im.shape[j] - dim - 1
            ri = np.random.randint(0, max_)
            dim_randints.append(ri)
        # Get the image chunk
        s_ = [slice(dim_randints[j], dim_randints[j] + shape[j]) for j in range(len(shape))]
        s_ = tuple(s_)
        x = im[s_]
        mbrightprop = x.mean() / x.max()
        if mbrightprop > min_brightness_prop: # if the image is bright enough
            # add the image slice to the slices dict (to be used later for GT segmentation)
            slices.append(s_)
            # add coords to output df
            for j in range(len(shape)):
                _add_to_dataframe(j, dim_randints[j], df)
            # add the image to the list of output chunks 
            # - note that these have yet to be augmented 
            xs.append(x)
            # get the datetime to give the samples unique names 
            now = datetime.now()
            d = now.strftime("%y%m%d_%H%M%S") + f'_img-{image_no}_chunk-{i}'
            ids.append(d)
            # another successful addition, job well done you crazy mofo
            i += 1
    df['data_ids'] = ids
    chunk_dict = {
        'x' : xs, 
        'slices' : slices, 
        'ids' : ids,
        'df' : df,
        'n' : len(xs)
    }
    chunk_dict['df'] = pd.DataFrame(chunk_dict['df'])
    return chunk_dict

# referenced in get_image_chunks(...)
def _add_to_dataframe(dim, start, df):
    if dim == 0:
        df['z_start'].append(start)
    if dim == 1:
        df['y_start'].append(start)
    if dim == 2:
        df['x_start'].append(start)


# ------------------------------
# Get Training Labels for Chunks
# ------------------------------

def get_labels_chunks(
    chunk_dict,
    ground_truth, 
    channels=('z-1', 'y-1', 'x-1', 'centreness-log'), 
    scale=(4, 1, 1), 
):  
    if not isinstance(channels, dict):
        channels = {'y' : channels}
    labels = {}
    #print(channels)
    chunk_dict['channels'] = channels
    for key in channels.keys():
        labs = get_training_labels(ground_truth, channels[key], scale)
        labels[key] = labs
    chunk_dict['ys'] = {}
    for key in labels.keys():
        chunk_dict['ys'][key] = []
    chunk_dict['ground_truth'] = []
    for s_ in chunk_dict['slices']:
        # add the segmentation chunks
        gt = ground_truth[s_]
        chunk_dict['ground_truth'].append(gt)
        # add the training labels chunks
        new_s_ = [slice(None, None), ] + list(s_)
        new_s_ = tuple(new_s_)
        for key in labels.keys():
            labs = labels[key]
            y = labs[new_s_]
            chunk_dict['ys'][key].append(y)
    return chunk_dict


# --------------
# Augment Chunks
# --------------
# Apply augmentation to each chunk. The augmentations are random and are 
# applied to each image, the ground truth, and the training labels. 
# If multiple sets of training labels are used, the same augmentation
# is applied to each. 

def augment_chunks(chunk_dict):
    x, ys, labs_keys, gt, n = _read_chunk_dict(chunk_dict)
    # augment each chunk overwriting memory as you go
    for i in range(n):
        labels_dict = {key: ys[key][i] for key in labs_keys}
        image, labels_dict, ground_truth = augment_images(x[i], labels_dict, gt[i])
        chunk_dict['x'][i] = image
        for key in labs_keys:
            chunk_dict['ys'][key][i] = labels_dict[key]
        chunk_dict['ground_truth'][i] = ground_truth
    return chunk_dict


def _read_chunk_dict(chunk_dict):
    x = chunk_dict['x'] # list (len n)
    ys = chunk_dict['ys'] # dict of lists (len n)
    labs_keys = list(ys.keys())
    gt = chunk_dict['ground_truth'] # list (len n)
    # check that all lists are len n
    n = chunk_dict['n']
    assert n == len(x)
    assert n == len(gt)
    for key in ys.keys():
        assert len(ys[key]) == n
    return x, ys, labs_keys, gt, n


# -----------
# Save Output
# -----------
# Saved output may be loaded at a later time for training (perhaps using 
# different trainin conditions), for reference, or used to test/visualise 
# quality of results. Saved output is generally not required for training 
# itself. 

def save_from_chunk_dict(chunk_dict, out_dir, name):
    x = chunk_dict['x'] # list (len n)
    ys = chunk_dict['ys'] # dict of lists (len n)
    labs_keys = list(ys.keys())
    channels = chunk_dict['channels']
    gt = chunk_dict['ground_truth'] # list (len n)
    ids = chunk_dict['ids']
    # get the folder to which to save the data
    chunk_dict['name'] = name
    # save the images and gt
    for i in range(len(x)):
        save_chunk(out_dir, i, x, ids, '_image.zarr')
        save_chunk(out_dir, i, gt, ids, '_GT.zarr')
    # make subdirectories for training labels and save
    labs_paths = {}
    for key in labs_keys:
        chans = channels[key]
        i = 0
        path = os.path.join(out_dir, str(key))
        labs_paths[key] = path
        os.makedirs(path, exist_ok=True)
        y = ys[key]
        for j in range(len(y)):
            save_chunk(path, i, y, ids, '_labels.zarr')
        i += 1
    chunk_dict['save_dir'] = out_dir
    chunk_dict['labels_dirs'] = labs_paths
    # convert ground truth to lazy rep so that it is accessible for vis 
    # but not using memory
    chunk_dict['ground_truth'] = da.stack(
        [da.from_array(chunk_dict['ground_truth'][i], 
         chunks=chunk_dict['ground_truth'][i].shape) \
        for i in range(len(chunk_dict['ground_truth']))])
    return out_dir
        

# -----------------------------
# Dask, Numpy, Torch Conversion
# -----------------------------

def convert_chunks_to_tensor(chunk_dict):
    x, ys, labs_keys, gt, n = _read_chunk_dict(chunk_dict)
    for i in range(n):
        chunk_dict['x'][i] = torch.from_numpy(x[i].copy())
        for key in labs_keys:
            chunk_dict['ys'][key][i] = torch.from_numpy(ys[key][i].copy())
        # chunk_dict['ground_truth'] = torch.from_numpy(gt[i].copy())
    return chunk_dict


def train_chunks_to_dask(train_dict):
    convert = ['x', 'vx', 'y', 'vy']
    for key in convert:
        stack = np.stack(train_dict[key])
        stack = da.from_array(stack, chunks=(1, ) + stack.shape[1:])
        train_dict[key] = stack
    return train_dict


def load_dask_as_tensor(i, stack):
    chunk = stack[i, ...].compute()
    chunk = torch.from_numpy(chunk)
    return chunk


def load_tensor_from_numpy(i, ls):
    chunk = ls[i].copy()
    chunk = torch.from_numpy(chunk)
    return chunk


def load_tensor_from_zarr(i, ls):
    chunk = np.array(ls[i])
    chunk = torch.from_numpy(chunk)
    return chunk

# ------------------------
# Concatenate chunk_dicts
# ------------------------

def concat_chunk_dicts(chunks_dict_list):
    # assume that the dicts have the right structure
    c = 0
    gt = []
    for chunk_dict in chunks_dict_list:
        if c == 0:
            full_dict = chunk_dict
            gt.append(chunk_dict['ground_truth'])
            c += 1
        else:
            full_dict['x'] = full_dict['x'] + chunk_dict['x']
            gt.append(chunk_dict['ground_truth'])
            full_dict['ids'] = full_dict['ids'] + chunk_dict['ids']
            for key in full_dict['ys'].keys():
                full_dict['ys'][key] = full_dict['ys'][key] + chunk_dict['ys'][key]
            full_dict['df'] = pd.concat([full_dict['df'], chunk_dict['df']])
            full_dict['n'] = full_dict['n'] + chunk_dict['n']
    gt = da.concatenate(gt, axis=0)
    full_dict['ground_truth'] = gt
    return full_dict




# ---------------------
# Convert to train_dict
# ---------------------

def chunk_dict_to_train_dict(chunk_dict, validation_prop=0.2):
    '''
    Will convert a chunk_dict into as many train_dicts as there
    are sets of labels to train on. 
    '''
    n = len(chunk_dict['x'])
    no_val = np.round(validation_prop * n).astype(int)
    vx_idx = np.random.randint(0, n, size=no_val)
    out = {}
    for key in chunk_dict['ys'].keys():
        train_dict = {
            'x' : [x for i, x in enumerate(chunk_dict['x']) if i not in vx_idx], 
            'vx' : [x for i, x in enumerate(chunk_dict['x']) if i in vx_idx], 
            'y' : [y for i, y in enumerate(chunk_dict['ys'][key]) if i not in vx_idx],
            'vy' : [y for i, y in enumerate(chunk_dict['ys'][key]) if i in vx_idx], 
            'ids' : [ID for i, ID in enumerate(chunk_dict['ids']) if i not in vx_idx],
            'vids' : [ID for i, ID in enumerate(chunk_dict['ids']) if i in vx_idx], 
            'out_dir' : chunk_dict['labels_dirs'][key], 
            'name' : key, 
            'channels' : chunk_dict['channels'][key]
        }
        #train_dict = train_chunks_to_dask(train_dict)
        print(f'generated train dict for {key}')
        out[key] = train_dict
    return out


# ------------------------
# General Helper Functions
# ------------------------

def normalise_data(image):
    '''
    Bring image values to between 0-1

    Parameters
    ----------
    image: np.array
        Image data. Dtype should be float.
    '''
    im = image / image.max()
    return im


def save_chunk(out_dir, i, data_list, ID_list, type_suffix):
    ID = ID_list[i]
    data = data_list[i]
    name = ID + type_suffix
    path = os.path.join(out_dir, name)
    zarr.save(path, data)
    # want to use out of memory zarr rather than in memory array
    arr = zarr.open_array(path, 'r')
    data_list[i] = arr


# -----------------------------------------------------------------------------
# Load Train Data
# -----------------------------------------------------------------------------

def load_train_data(
    data_dir, 
    id_regex=r'\d{6}_\d{6}_\d{1,3}',
    x_suffix='_image.tif', 
    y_suffix='_labels.tif',
    gt_suffix='_GT.tif',
    load_GT=False
    ):
    pass


def load_train_data(
                    data_dir, 
                    id_regex=r'\d{6}_\d{6}_\d{1,3}',
                    x_regex=r'\d{6}_\d{6}_\d{1,3}_image.tif', 
                    y_regex=r'\d{6}_\d{6}_\d{1,3}_labels.tif'
                    ):
    '''
    Load train data from a directory according to a naming convention

    Parameters
    ----------
    data_dir: str
        Directory containing data
    id_regex: r string
        regex that will be used to extract IDs that will
        be used to label network output
    x_regex: r string
        regex that represents image file names, complete with
        extension (tiff please)
    y_regex: r string
        regex that represents training label files, complete
        with extension (tiff please)

    Returns
    -------
    xs: list of torch.Tensor
        List of images for training 
    ys: list of torch.Tensor
        List of affinities for training
    ids: list of str
        ID strings by which each image and label are named.
        Eventually used for correctly labeling network output
    
    '''
    # Get file names for images and training labels
    x_paths, y_paths = get_files(
                                 data_dir, 
                                 x_regex=x_regex, 
                                 y_regex=x_regex
                                 )
    # Get IDs 
    id_pattern = re.compile(id_regex)
    ids = []
    x_paths.sort()
    y_paths.sort()
    for i in range(len(x_paths)):
        xn = Path(x_paths[i]).stem # this could have been avoided
        yn = Path(y_paths[i]).stem # why would I bother now though?!
        # assumes there will be a match for each
        xid = id_pattern.search(xn)[0]
        yid = id_pattern.search(yn)[0]
        m = 'There is a mismatch in image and label IDs'
        assert xid == yid, m
        ids.append(xid)
    # Get images and training labels in tensor form 
    xs = []
    ys = []
    for i in range(len(x_paths)):
        xp = x_paths[i]
        yp = y_paths[i]
        x = imread(xp)
        x = normalise_data(x)
        y = imread(yp)
        xs.append(torch.from_numpy(x))
        ys.append(torch.from_numpy(y))
    # returns objects in the same manner as get_train_data()
    print('------------------------------------------------------------')
    print(f'Loaded {len(xs)} sets of training data')
    print_labels_info(ys[0].shape)
    return xs, ys, ids


if __name__ =="__main__":
    import zarr
    import napari
    # Directory for training data and network output 
    #data_dir = '/Users/amcg0011/Data/pia-tracking/cang_training'
    data_dir = '/home/abigail/data/platelet-segmentation-training'
    # Path for original image volumes for which GT was generated
    image_paths = [os.path.join(data_dir, '191113_IVMTR26_I3_E3_t58_cang_training_image.zarr')] 
    # Path for GT labels volumes
    labels_paths = [os.path.join(data_dir, '191113_IVMTR26_I3_E3_t58_cang_training_labels.zarr')]
    labs = zarr.open(labels_paths[0])
    labs = np.array(labs)
    #cent_off = get_centre_offsets(labs, (4, 1, 1))
    #v = napari.Viewer()
    #z = cent_off[0] # - cent_off[0].min()
    #v.add_image(z, name='Z offsets', colormap='bop purple', blending='additive', scale=(4, 1, 1))
    #y = cent_off[1] #- cent_off[1].min()
    #v.add_image(y, name='Y offsets', colormap='bop orange', blending='additive', scale=(4, 1, 1))
    #x = cent_off[2] #- cent_off[2].min()
    #v.add_image(x, name='X offsets', colormap='bop blue', blending='additive', scale=(4, 1, 1))
    #napari.run()



import dask.array as da
from dask import delayed
import numpy as np
import os
from pathlib import Path
import re
import skimage.io as io


LINE = '------------------------------------------------------------'

def get_files(
              data_dir, 
              x_regex=r'\d{6}_\d{6}_\d{1,3}_image.tif', 
              y_regex=r'\d{6}_\d{6}_\d{1,3}_labels.tif'
              ):
    '''
    '''
    files = os.listdir(data_dir)
    x_paths = get_paths(
                        data_dir, 
                        regex=x_regex, 
                        )
    y_paths = get_paths(
                        data_dir, 
                        regex=y_regex, 
                        )
    m = 'There is a mismatch in the number of images and training labels'
    assert len(x_paths) == len(y_paths), m
    return x_paths, y_paths


def get_paths(
              data_dir, 
              regex=r'\d{6}_\d{6}_\d{1,3}_output.tif'
              ):
    '''
    Awesome default for regex... just so lazy !!
    '''
    files = os.listdir(data_dir)
    paths = []
    pattern = re.compile(regex)
    for f in files:
        match = pattern.search(f)
        if match is not None:
            paths.append(os.path.join(data_dir, match[0]))
    return paths


def write_log(string, out_dir, log_name='log.txt'):
    os.makedirs(out_dir, exist_ok=True)
    with open(os.path.join(out_dir, log_name), 'a') as log:
        log.write(string + '\n')


def log_dir_or_None(log, out_dir):
    if log:
        return out_dir
    else:
        return None

def get_ids(paths, regex=r'\d{6}_\d{6}_\d{1,3}'):
    ids = []
    pattern = re.compile(regex)
    for p in paths:
        p = Path(p)
        name = p.stem
        match = pattern.search(name)
        if match is not None:
            ids.append(match[0])
        else:
            raise ValueError('Irregular ID for training data file: must be YYMMDD_HHMMSS_<digit>')
    return ids


def check_ids_match(x, y, regex=r'\d{6}_\d{6}_\d{1,3}'):
    pattern = re.compile(regex)
    m = f'Number of '
    assert len(x) == len(y), m
    for i in range(len(x)):
        if not os.path.exists(x[i]): # if just an id string
            assert x[i] == y[i]
        else: # if the ids to be matched are in path stems
            xn = Path(x).stem
            yn = Path(y).stem
            xid = pattern.search(xn)[0] # will raise error if no id found
            yid = pattern.search(yn)[0]
            assert xid == yid


# -------------------------
# Load Output from Training 
# -------------------------

def get_dataset(train_dir, out_dir=None, GT=False, validation=False, return_ID=False):
    # directory for output, if none, assume output is with training data
    if out_dir is None:
        out_dir = train_dir
    # Get output regular expression
    if validation == False:
        out_regex = r'\d{6}_\d{6}_\d{1,3}_output.tif'
        o_s = '_output.tif'
    else:
        out_regex = r'\d{6}_\d{6}_\d{1,3}_validation_output.tif'
        o_s = '_validation_output.tif'
    # Get output paths and IDs
    output_paths = sorted(get_paths(out_dir, out_regex))
    ids = get_ids(output_paths)
    #output = get_regex_images(out_dir, out_regex, ids)
    # get training data (images and labels) according to id strings, 
    #   which correspond to the batch number.
    suffixes = ['_image.tif', '_labels.tif', o_s]
    if GT:
        suffixes.append('_GT.tif')
        labs, images, output, ground_truth =  get_data_by_id(train_dir, suffixes, 
                                       out_dir=out_dir, validation=validation)
        if not return_ID:
            return images, labs, output, ground_truth
        else:
            return images, labs, output, ground_truth, ids
    else:
        labs, images, output =  get_data_by_id(train_dir, suffixes, 
                                       out_dir=out_dir, validation=validation)
        if not return_ID:
            return images, labs, output
        else:
            return images, labs, output, ids


def get_dataset_segs(train_dir, out_dir=None, validation=True):
    suffixes = ('_GT.tif', '_segmentation.tif', '_DoG-segmentation.tif', '_image.tif')
    gt, seg, dog, im = get_data_by_id(train_dir, suffixes, 
                              out_dir=out_dir, validation=validation)
    return gt, seg, dog, im


def get_data_by_id(train_dir, suffixes, out_dir=None, validation=False):
    # directory for output, if none, assume output is with training data
    if out_dir is None:
        out_dir = train_dir
    # Get output regular expression
    if not validation:
        out_regex = r'\d{6}_\d{6}_\d{1,3}_output.tif'
    else:
        out_regex = r'\d{6}_\d{6}_\d{1,3}_validation_output.tif'
    # Get output paths and IDs
    output_paths = sorted(get_paths(out_dir, out_regex))
    ids = get_ids(output_paths)
    id_regex = r'\d{6}_\d{6}_\d{1,3}'
    out = []
    for s in suffixes:
        regex = id_regex + s
        out.append(get_regex_images(train_dir, regex, ids))
    return tuple(out)


def get_regex_images(data_dir, regex, ids, id_regex=r'\d{6}_\d{6}_\d{1,3}'):
    id_pattern = re.compile(id_regex)
    imread = delayed(_imread_squeeze, pure=True)  
    file_paths = sorted(get_paths(data_dir, regex))
    correct_paths = []
    for ID in ids:
        id_done = False
        for f in file_paths:
            n = Path(f).stem
            id_match = id_pattern.search(n)[0] # assumes there will be one
            if id_match == ID:
                correct_paths.append(f)
                id_done = True
        m = f'No file match was found for ID: {ID}'
        assert id_done, m
    images = [imread(path) for path in correct_paths]  
    sample = images[0].compute()  
    arrays = [da.from_delayed(image,           
                              dtype=sample.dtype,   
                              shape=sample.shape)
                                for image in images]

    stack = da.stack(arrays, axis=0)
    return stack


def _imread_squeeze(path):
    img = io.imread(path)
    return np.squeeze(img)


from matplotlib.colors import same_color
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import os
from pathlib import Path
import re
from .plots import VI_plot, plot_AP, plot_count_difference

import pandas as pd
from skimage.measure import regionprops
from skimage.metrics import variation_of_information
import umetrix

import napari
from scipy import stats

# compare model result with ground truth validation data (not seen by network)

# SEG METRICS
#   - Average precision
#   - Variation of information
#   - Object count difference


# plot accuracy metrics against the following to analyse where models fail
#   because #visualanalyticsbro

# IM METRICS
#   - image entropy
#   - average intensity

# OB METICS
#   - average volume
#   - density
#   - average elongation/flattness



# --------------------
# Segmentation Metrics
# --------------------


def get_accuracy_metrics(
    slices, 
    gt_data: napari.layers.Labels,
    model_result:  napari.layers.Labels,
    name: str,
    prefix : str,
    VI: bool = True, 
    AP: bool = True, 
    ND: bool = True,
    out_path = None,
    exclude_chunks: int = 10, 
    ):
    '''
    Parameters:
    slices: list of tupel of 4 x slice 
        Slices denoting where the image chunks to be taken as single units of
        network output end and begin.
    gt_data: napari.types.LabelsData or int array like
        Stack of N x 3D validation ground truth (i.e., 4D with N,z,y,x)
    model_result: napari.types.LabelsData or int array like
        Stack of N x 3D model derived segmentation (i.e., 4D with N,z,y,x)
    VI: bool
        Should we find variation of information scores
    AP: bool
        Should we find average precision scores (for IoU 0.6-0.9)
    ND: bool
        SHould we find the number of objects difference from 
    '''
    scores = {
        'VI: GT | Output' : [], 
        'VI: Output | GT' : [], 
        'Number objects (GT)' : [], 
        'Number objects (model)' : [], 
        'Count difference' : [], 
        'Count difference (%)' : [], 
        }
    IoU_dict = generate_IoU_dict()
    scores.update(IoU_dict)
    if isinstance(gt_data, napari.layers.Labels):
        gt_data = gt_data.data
    if isinstance(model_result, napari.layers.Labels):
        model_result = model_result.data
    if gt_data.ndim != model_result.ndim:
        dim_dif = gt_data.ndim - model_result.ndim
        if dim_dif == -1:
            # if 3D and 4D will assume 3D needs to be replicated t times (data is tzyx) and stacked
            gt_data = [gt_data, ] * model_result.shape[0]
            gt_data = np.stack(gt_data)
        elif dim_dif == 1:
            model_result = [model_result, ] * gt_data.shape[0]
            model_result = np.stack(model_result)
        else:
            raise ValueError('Ground truth and model result must be either 3D or 4D arrays')
    for s_, c_ in slices:
        gt = gt_data[s_]
        gt = np.squeeze(gt)[c_]
        n_objects = np.unique(gt).size
        if n_objects > exclude_chunks + 1:
            mr = model_result[s_]
            mr = np.squeeze(mr)[c_]
            #print('n_objects', n_objects)
            if VI:
                vi = variation_of_information(gt, mr)
                scores['VI: GT | Output'].append(vi[0])
                scores['VI: Output | GT'].append(vi[1])
            if AP:
                generate_IoU_data(gt, mr, scores)
            if ND:
                n_mr = np.unique(mr).size
                #print('n_mr', n_mr, np.unique(mr), mr.shape, mr.dtype)
                nd = n_mr - n_objects
                ndp = nd / n_objects * 100 # as a percent might be more informative
                scores['Count difference (%)'].append(ndp)
                scores['Number objects (GT)'].append(n_objects)
                scores['Number objects (model)'].append(n_mr)
                scores['Count difference'].append(nd)
    lens = {key : len(scores[key]) for key in scores.keys()}
    to_keep = [key for key in scores.keys() if lens[key] > 1]
    new_scores = {key : scores[key] for key in to_keep}
    new_scores = pd.DataFrame(new_scores)
    statistics = single_sample_stats(new_scores, to_keep, name)
    new_scores['model_name'] = [name, ] * len(new_scores)
    if out_path is not None:
        n = prefix + '_' + name + '_scores.csv'
        scores_path = os.path.join(out_path, n)
        new_scores.to_csv(scores_path)
        statistics = statistics.T
        n = prefix + '_' + name + '_stats.csv'
        stats_p = os.path.join(out_path, n)
        statistics.to_csv(stats_p)
    ap_scores = None
    if AP:
        ap_scores = generate_ap_scores(new_scores, name)
        if out_path is not None:
            n = prefix + '_' + name + '_AP_curve.csv'
            AP_path = os.path.join(out_path, n)
            ap_scores.to_csv(AP_path)
    return (new_scores, ap_scores), statistics


def single_sample_stats(df, columns, name):
    results = {}
    alpha = 0.95
    for c in columns:
        sample_mean = np.mean(df[c].values)
        sample_sem = stats.sem(df[c].values)
        degrees_freedom = df[c].values.size - 1
        CI = stats.t.interval(alpha, degrees_freedom, sample_mean, sample_sem)
        n = str(c) + '_'
        results[n + 'mean'] = [sample_mean, ]
        results[n + 'sem'] = [sample_sem, ]
        results[n + '95pcntCI_2-5pcnt'] = [CI[0], ]
        results[n + '95pcntCI_97-5pcnt'] = [CI[1], ]
    results = pd.DataFrame(results)
    results['model_name'] = name
    return results


#def metrics_for_stack(directory, name, seg, gt):
#    assert seg.shape[0] == gt.shape[0]
#    IoU_dict = generate_IoU_dict()
#    for i in range(seg.shape[0]):
#        seg_i = seg[i].compute()
#        gt_i = gt[i].compute()
#        generate_IoU_data(gt_i, seg_i, IoU_dict)
#    df = save_data(IoU_dict, name, directory, 'metrics')
#    ap = generate_ap_scores(df, name, directory)
#    return df, ap


def calc_ap(result):
        denominator = result.n_true_positives + result.n_false_negatives + result.n_false_positives
        return result.n_true_positives / denominator


def generate_IoU_dict(thresholds=(0.3, 0.35, 0.4, 0.45, 0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9)):
    IoU_dict = {}
    IoU_dict['n_predicted'] = []
    IoU_dict['n_true'] = []
    IoU_dict['n_diff'] = []
    for t in thresholds:
        n = f't{t}_true_positives'
        IoU_dict[n] = []
        n = f't{t}_false_positives'
        IoU_dict[n] = []
        n = f't{t}_false_negatives'
        IoU_dict[n] = []
        n = f't{t}_IoU'
        IoU_dict[n] = []
        n = f't{t}_Jaccard'
        IoU_dict[n] = []
        n = f't{t}_pixel_identity'
        IoU_dict[n] = []
        n = f't{t}_localization_error'
        IoU_dict[n] = []
        n = f't{t}_per_image_average_precision'
        IoU_dict[n] = []
    return IoU_dict


def generate_IoU_data(gt, seg, IoU_dict, thresholds=(0.3, 0.35, 0.4, 0.45, 0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9)):
    for t in thresholds:
        result = umetrix.calculate(gt, seg, strict=True, iou_threshold=t)
        n = f't{t}_true_positives'
        IoU_dict[n].append(result.n_true_positives) 
        n = f't{t}_false_positives'
        IoU_dict[n].append(result.n_false_positives) 
        n = f't{t}_false_negatives'
        IoU_dict[n].append(result.n_false_negatives) 
        n = f't{t}_IoU'
        IoU_dict[n].append(result.results.IoU) 
        n = f't{t}_Jaccard'
        IoU_dict[n].append(result.results.Jaccard) 
        n = f't{t}_pixel_identity'
        IoU_dict[n].append(result.results.pixel_identity) 
        n = f't{t}_localization_error'
        IoU_dict[n].append(result.results.localization_error) 
        n = f't{t}_per_image_average_precision'
        IoU_dict[n].append(calc_ap(result))
        if t == thresholds[0]:
            IoU_dict['n_predicted'].append(result.n_pred_labels)
            IoU_dict['n_true'].append(result.n_true_labels)
            IoU_dict['n_diff'].append(result.n_true_labels - result.n_pred_labels)


def save_data(data_dict, name, prefix, directory, suffix):
    df = pd.DataFrame(data_dict)
    n = prefix + '_' + name + '_' + suffix +'.csv'
    p = os.path.join(directory, n)
    df.to_csv(p)
    return df


def generate_ap_scores(
        df, 
        name,  
        thresholds=(0.3, 0.35, 0.4, 0.45, 0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9)
        ):
    ap_scores = {'average_precision' : [], 
                 'threshold': []}
    for t in thresholds:
        ap_scores['threshold'].append(t)
        n = f't{t}_true_positives'
        true_positives = df[n].sum()
        n = f't{t}_false_positives'
        false_positives = df[n].sum()
        n = f't{t}_false_negatives'
        false_negatives = df[n].sum()
        ap = true_positives / (true_positives + false_negatives + false_positives)
        ap_scores['average_precision'].append(ap)
    ap_scores['model_name'] = [name, ] * len(thresholds)
    ap_scores = pd.DataFrame(ap_scores)
    #print(ap_scores)
    return ap_scores



# ----------------
# Plotting Methods
# ----------------
def plot_accuracy_metrics(
    data: tuple,
    prefix: str,
    save_dir: str,
    name: str,
    variation_of_information: bool, 
    average_precision: bool, 
    object_count: bool,
    show: bool =True
    ):
    '''
    Parameters
    ----------
    data: tuple of pd.DataFrame
        df0 w/ VI/IoU/nd data, df1 w/ AP data
    prefix: str
        name under which to 
    save_dir: str
        directory into which to save plots
    show: bool
        show the plots?
    '''
    df0 = data[0]
    df1 = data[1]
    if variation_of_information:
        n = prefix + '_' + name + '_VI_plot.pdf'
        VI_path = os.path.join(save_dir, n)
        VI_plot(
                df0, 
                cond_ent_over='VI: GT | Output', 
                cond_ent_under='VI: Output | GT', 
                save=VI_path, show=show)
    if average_precision:
        n = prefix + '_' + name + '_AP_plot.pdf'
        AP_path = os.path.join(save_dir, n)
        df1 = [df1, ]
        plot_AP(df1, [prefix, ], AP_path, 'Average precision', show=show)
    if object_count:
        n = prefix + '_' + name + '_OD_plot.pdf'
        OD_path = os.path.join(save_dir, n)
        plot_count_difference(df0, 'Object count difference', OD_path, show=show)



# ---------
# OLD STUFF
# ---------

def _get_index(chans, opts):
    idx = []
    for opt in opts:
        for i, c in enumerate(chans):
            if c == opt:
                idx.append(i)
    if len(idx) == 0:
        raise ValueError(f'No channel in {chans} matches the options {opts}')
    return idx[0]


def _get_VI_paths(data_dir, seg_info, validation=False):
    if validation:
        s = 'validation_VI.csv'
    else:
        s = '_VI.csv'
    paths, names = _get_data_paths(data_dir, seg_info, s)
    return paths, names


def _get_AP_paths(data_dir, seg_info, validation=True):
    if validation:
        s = '_validation_AP.csv'
    else:
        s = '_test_AP.csv'
    paths, names = _get_data_paths(data_dir, seg_info, s)
    return paths, names
        

def _get_IoU_paths(data_dir, seg_info, validation=True):
    if validation:
        s = '_validation_metrics.csv'
    else:
        s = '_test_metrics.csv'
    paths, names = _get_data_paths(data_dir, seg_info, s)
    return paths, names


def _get_data_paths(data_dir, seg_info, s):
    paths = []
    names = []
    for key in seg_info.keys():
        dir_ = seg_info[key]['directory']
        suffix = seg_info[key]['suffix']
        p = os.path.join(dir_, suffix + s)
        if not os.path.exists(p):
            raise ValueError(f'Cannot find path {p}')
        paths.append(p)
        n = seg_info[key]['suffix']
        names.append(n)
    return paths, names



# -------------
# Image Metrics
# -------------


def get_image_metrics():
    pass


# --------------------------
# Experimental image metrics
# --------------------------

def affinity_sum_graph(img, affs=(1, 2, 3, 5, 10, 20, 40)):
    results = []
    dims = len(img.shape)
    for a in affs:
        sums = []
        for ax in range(dims):
            diff = np.diff(img, n=a, axis=ax)
            norm_sum = np.abs(np.sum(diff)/diff.size)
            sums.append(norm_sum)
        results.append(np.sum(sums))
    return list(affs), results
            


# --------------
# Object Metrics
# --------------


def get_object_metrics():
    pass



import ast
from typing import Union
import numpy as np
import napari
from magicgui import magic_factory
from .predict import make_chunks
from .training_experiments import get_experiment_dict, run_experiment
import zarr
from skimage.io import imread
import os
from .metrics import get_accuracy_metrics, plot_accuracy_metrics
from .plots import comparison_plots
from typing import Union
import json
from pathlib import Path
from .segmentation import segmenters
import dask.array as da
from dask import delayed


# ------------
# Train widget
# ------------


@magic_factory(
    call_button=True, 
    mask_prediction={'choices': ['mask', 'centreness']}, 
    centre_prediciton={'choices': ['centreness-log', 'centreness', 'centroid-gauss']},
    affinities_extent={'widget_type' : 'LiteralEvalLineEdit'},
    training_name={'widget_type': 'LineEdit'}, 
    loss_function={'choices': ['BCELoss', 'DiceLoss']}, 
    output_dir={'widget_type': 'FileEdit', 'mode' : 'd'}, 
    scale={'widget_type' : 'LiteralEvalLineEdit'},
    learning_rate={'widget_type' : 'LiteralEvalLineEdit'}
    )
def train_from_viewer(
    viewer: napari.viewer.Viewer, 
    image_stack: napari.layers.Image, 
    labels_stack: napari.layers.Labels,
    output_dir: Union[str, None]=None, 
    scale: tuple=(1, 1, 1), 
    mask_prediction='mask', 
    centre_prediciton='centreness-log', #lol btw this is a typo in the whole repo :P ... centredness 
    affinities_extent=1, 
    training_name='my-unet',
    loss_function='BCELoss', 
    learning_rate=0.01, 
    epochs=4,
    validation_prop=0.2, 
    n_each=50,
    predict_labels: bool=True,
    save_labels=True,
    ):
    '''
    Train a U-net from the viewer. This will ideally be updated in later releases to allow for other network archetectures/segmentation algorithms to be trained (more in the fashion of segment_data). 

    Parameters
    ----------
    viewer: napari.viewer.Viewer

    image_stack: napari.layers.Image 
    output_dir: str or None (None)
    scale: tuple (1, 1, 1) 
    mask_prediction: str ('mask')
    centre_prediciton: str ('centreness-log')
    affinities_extent: int (1)
    training_name: str ('my-unet')
    loss_function: str ('BCELoss') 
    learning_rate: scalar (0.01)
    epochs: int (4)
    validation_prop: (0.2)
    n_each: (50)
    predict_labels: bool (True)
    save_labels: (True)
    '''
    _train_from_viewer(viewer, image_stack, labels_stack, output_dir, scale, 
        mask_prediction, centre_prediciton, affinities_extent, training_name, 
        loss_function, learning_rate, epochs, validation_prop, n_each, predict_labels, save_labels)


def _train_from_viewer(
    viewer: napari.viewer.Viewer, 
    image_stack: napari.layers.Image, 
    labels_stack: napari.layers.Labels,
    output_dir: Union[str, None]=None, 
    scale: tuple=(1, 1, 1), 
    mask_prediction='mask', 
    centre_prediciton='centreness-log', #lol btw this is a typo in the whole repo :P
    affinities_extent=1, 
    training_name='my-unet',
    loss_function='BCELoss', 
    learning_rate=0.01, 
    epochs=4,
    validation_prop=0.2, 
    n_each=50,
    predict_labels: bool=True,
    save_labels: bool=True,
    ):

    # Prepare training data
    # ---------------------
    if isinstance(image_stack, napari.layers.Image):
        image_4D_stack = image_stack.data
    else:
        image_4D_stack = image_stack

    if isinstance(labels_stack, napari.layers.Labels):
        labels_4D_stack = labels_stack.data
    else:
        labels_4D_stack = labels_stack
    # in case there is an accidental extra dim
    image_4D_stack = np.squeeze(image_4D_stack)
    labels_4D_stack = np.squeeze(labels_4D_stack)
    # make sure the image and gt are the same shape
    assert image_4D_stack.shape == labels_4D_stack.shape
    # if the image is 3D convert to 4D
    if image_4D_stack.ndim == 3:
        image_4D_stack = image_4D_stack[np.newaxis, :, :, :]
        labels_4D_stack = labels_4D_stack[np.newaxis, :, :, :]
    condition_name = [training_name, ]
    image_list = [image_4D_stack[i, ...] for i in range(image_4D_stack.shape[0])]
    labels_list = [labels_4D_stack[i, ...] for i in range(labels_4D_stack.shape[0])]
    del image_4D_stack
    del labels_4D_stack

    # Construct training info
    # -----------------------
    channels_list = construct_channels_list(affinities_extent, mask_prediction, 
                                        centre_prediciton)
    print('Channels list:')
    print(channels_list)
    conditions_list = construct_conditions_list(image_list, loss_function, 
                                                learning_rate, epochs, scale)
    print('Conditions list:')
    print(conditions_list)
    exp_dict = get_experiment_dict(channels_list, condition_name, 
                                   conditions_list=conditions_list, 
                                   validation_prop=validation_prop, 
                                   n_each=n_each)
    print('Experiment dict:')
    print(exp_dict)

    # Run training experiment
    # -----------------------
    u_path = run_experiment(exp_dict, image_list, labels_list, output_dir)

    # Predict full labels
    # -------------------
    if predict_labels:
        if save_labels:
            save_path = os.path.join(str(output_dir), training_name + '_labels-prediction.zarr')
        else:
            save_path = None
        seg_func = segmenters['affinity-unet-watershed']
        seg_func(napari_viewer=viewer, input_volume_layer=image_stack, save_dir=output_dir, 
             name=f'{training_name}_labels', unet_or_config_file=u_path[0], layer_reference=None, 
              chunk_size=(10, 256, 256), margin=(1, 64, 64), debug=False)
        labels_layer = viewer.layers[f'{training_name}_labels']
    # Save metadata
    # -------------
    meta = {
        'unet' : u_path[0], 
        'chunk_size' : (10, 256, 256), 
        'margin' : (1, 64, 64), 
        'mask_prediction' : mask_prediction, 
        'centre_prediction' : centre_prediciton, 
        'affinities_extent' : affinities_extent, 
        'loss_function' : loss_function, 
        'output_dir' : str(output_dir), 
        'learning_rate' : learning_rate, 
        'epochs' : epochs, 
        'validation_prop' : validation_prop, 
        'n_each' : n_each, 
        'labels_path' : save_path
    }
    if predict_labels and isinstance(labels_layer, napari.layers.Labels):
        labels_layer.metadata.update(meta)
    json_object = json.dumps(meta, indent=4)
    meta_path = os.path.join(output_dir, Path(u_path[0]).stem + '_meta.json')
    with open(meta_path, "w") as outfile:
        outfile.write(json_object)



def construct_channels_list(
    affinities_extent, 
    mask_prediction, 
    centre_predicition
    ):
    dims = ('z', 'y', 'x')
    affs = []
    if isinstance(affinities_extent, tuple):
        m = f'please ensure the length of the affinities extent tuple matches the number of dims in {dims}'
        assert len(affinities_extent) == len(dims), m
    elif isinstance(affinities_extent, int):
        affinities_extent = [affinities_extent, ] * len(dims)
        affinities_extent = tuple(affinities_extent)
    else:
        m = 'Please insert affinities extent of type tuple or int (e.g., 1 or (2, 2, 1))'
        raise TypeError(m)
    for i, d in enumerate(dims):
        n_affs = affinities_extent[i]
        for n in range(1, n_affs + 1):
            affs.append(f'{d}-{n}')
    affs.append(mask_prediction)
    affs.append(centre_predicition)
    affs = [tuple(affs), ]
    return affs


def construct_conditions_list(
    image_list, 
    loss_function, 
    learning_rate, 
    epochs, 
    scale
    ):
    scale = [scale for l in image_list]
    condition_dict = {
        'scale' : scale, 
        'lr' : learning_rate, 
        'loss_function' : loss_function, 
        'epochs' : epochs
    }
    return [condition_dict, ]


# -------------------------
# Load Image or Labels Data
# -------------------------

@magic_factory(
    directory={'widget_type': 'FileEdit', 'mode' : 'd'}, 
    data_file={'widget_type': 'FileEdit'},
    data_type={'choices': ['individual frames', 'image stacks']},
    layer_name={'widget_type' : 'LineEdit'},
    layer_type={'choices': ['Image', 'Labels', 'Shapes']},
    scale={'widget_type' : 'LiteralEvalLineEdit'}, 
    translate={'widget_type' : 'LiteralEvalLineEdit'}, 
)
def load_data(
    napari_viewer: napari.viewer.Viewer, 
    layer_name:str,
    layer_type: str='Image',
    data_type: str ='individual frames',
    directory: Union[str, None] =None, 
    data_file: Union[str, None] =None,
    scale: tuple =(1, 1, 1),
    translate: tuple =(0, 0, 0),
    split_channels: bool=False, 
    in_memory: bool=True,
    ):
    '''
    Load the data into the viewer as a stack of 3D image frames. 

    Parameters
    ----------
    layer name: str (None)
        What will you call the layer of images you want to load in?
    layer_type: str ('Image')
        Is the data you want to load a segmentation (use "Labels") or an image (use "Image")
    data_type: str ('individual frames')
        Are the data a series of individual 3D images? If so use "individual images". If the 
        data are in the format of a single 4D or 5D image also use "individual images". If 
        you have a directory with several 4D images, use "image stacks". 
    directory:
        Path to the a directory containing image stacks (4D) or individual frames (3D) or
        to a zarr file. Data can be 3D (ZYX), 4D (TZYX), or 5D (CTZYX). Zarr files must
        end in .zarr/.zar. If you want to load a single tiff file, leave blank. 
    data_file: str (None)
        Path to the image stack (if data_type is "image stacks") or to 
        a single image frame (if data_type is "individual frames"). 
        This needs to end in .tiff/.tif. If you want to load from a directory or 
        want to load a zarr, leave blank. 
    scale: tuple of float
        Scale of the image/segmentation to display in z, y, x format.
    translate: tuple of float
        Translation of the image origin in z, y, x format.
    split_channels:
        If you have a multichannel image in the format CTZYX (or any format really as
        long as the first dim is the channel) you can split these. This is useful if 
        you are loading 5D data to segment. 
    in_memory:
        Do you want all of the data to be loaded into your computer's memory (i.e., 
        in RAM). Select false if your data is bigger than RAM or very large.
    save_stack:
        If you loaded a stack of images, do you want to save these as a single stack 
        of zarr files. If you do this for a stack of labels, you will be able to paint
        directly into the files when correcting labels. 
    '''
    _load_data(napari_viewer, layer_name, 
               layer_type, data_type,
                 directory, data_file,
                scale, translate, split_channels, 
                in_memory)


def _load_data(
    napari_viewer: napari.viewer.Viewer, 
    layer_name: str,
    layer_type: str,
    data_type: str ='individual frames',
    directory: Union[str, None] =None, 
    data_file: Union[str, None] =None,
    scale: tuple =(1, 1, 1),
    translate: tuple =(0, 0, 0),
    split_channels: bool=False, 
    in_memory: bool=True,
    ):
    '''
    Load the data into the viewer as a stack of 3D image frames. 

    Parameters
    ----------
    layer name: str (None)
        What will you call the layer of images you want to load in?
    layer_type: str ('Image')
        Is the data you want to load a segmentation (use "Labels") or an image (use "Image")
    data_type: str ('individual frames')
        Are the data a series of individual 3D images? If so use "individual images". If the 
        data are in the format of a single 4D or 5D image also use "individual images". If 
        you have a directory with several 4D images, use "image stacks". 
    directory:
        Path to the a directory containing image stacks (4D) or individual frames (3D) or
        to a zarr file. Data can be 3D (ZYX), 4D (TZYX), or 5D (CTZYX). Zarr files must
        end in .zarr/.zar. If you want to load a single tiff file, leave blank. 
    data_file: str (None)
        Path to the image stack (if data_type is "image stacks") or to 
        a single image frame (if data_type is "individual frames"). 
        This needs to end in .tiff/.tif. If you want to load from a directory or 
        want to load a zarr, leave blank. 
    scale: tuple of float
        Scale of the image/segmentation to display in z, y, x format.
    translate: tuple of float
        Translation of the image origin in z, y, x format.
    split_channels:
        If you have a multichannel image in the format CTZYX (or any format really as
        long as the first dim is the channel) you can split these. This is useful if 
        you are loading 5D data to segment. 
    in_memory:
        Do you want all of the data to be loaded into your computer's memory (i.e., 
        in RAM). Select false if your data is bigger than RAM or very large.
    save_stack:
        If you loaded a stack of images, do you want to save these as a single stack 
        of zarr files. If you do this for a stack of labels, you will be able to paint
        directly into the files when correcting labels. 
    '''
    if directory is not None:
        directory = str(directory)
    if data_file is not None:
        data_file = str(data_file)
    if layer_type == 'Image' or layer_type == 'Labels':
        imgs, uses_directory = read_data(directory, data_file, data_type, in_memory)
        if imgs.ndim > 3:
            if not split_channels:
                add_to_scale = (1, ) * (imgs.ndim - 3)
                add_to_translate = (0, ) * (imgs.ndim - 3)
            else: 
                add_to_scale = (1, ) * (imgs.ndim - 4)
                add_to_translate = (0, ) * (imgs.ndim - 4)
            scale = add_to_scale + scale
            translate = add_to_translate + translate
        if layer_type == 'Image':
            if not split_channels:
                napari_viewer.add_image(imgs, scale=scale, name=layer_name, translate=translate)
            else:
                for i in range(imgs.shape[0]):
                    napari_viewer.add_image(imgs[i, ...], scale=scale, name=layer_name, translate=translate)
        if layer_type == 'Labels':
            napari_viewer.add_labels(imgs, scale=scale, name=layer_name, translate=translate)
    if layer_type == 'Shapes':
        if data_file is not None:
            shapes = read_shapes(data_file)
        elif directory is not None:
            files = [os.path.join(directory, f) for f in os.listdir(directory) if f.endswith('.npy')]
            shapes = []
            for f in files:
                new = read_shapes(f)
                shapes = shapes + new
        else:
            raise ValueError("Please ensure you pick a file or directory to read from")
        napari_viewer.add_shapes(shapes, scale=scale, name=layer_name, translate=translate)


def read_shapes(data_file):
    shapes = np.load(str(data_file), allow_pickle=False)
    shapes = [s for s in shapes]
    return shapes

# TODO: refactor this **awful** function... you can do better
def read_data(directory, data_file, data_type, in_memory):
    """
    When supplied with a directory that is a zarr, this function will open
    the image as a zarr (not in memory). When supplied with a directory
    of files, the function will open any tiffs or zarrs in the directoy
    as a 4D stack of 3D images. When supplied with a single file that is
    a tiff, the tiff will be loaded (regardless of dimension). 

    If in_memory, the data will be opened as a numpy array. If this is 
    set to False, the data will be opened as a dask array. This is neccessary
    if your image is too big to fit in the computer's RAM. 
    """
    possible_suf = ['.zarr', '.zar', '.tiff', '.tif']
    # is the data coming from a directory (not a zarr file)
    uses_directory = directory is not None
    is_zarr = False
    if uses_directory:
        uses_directory = os.path.isdir(directory) and not directory.endswith('.zarr') and not directory.endswith('.zar')
    # is the data coming from a single file
    single_file = data_file is not None
    if single_file:
        if data_file.endswith('.tiff') or data_file.endswith('.tif'):
            data_paths = [data_file, ]
    elif not uses_directory:
        is_zarr = directory.endswith('.zarr') or directory.endswith('.zar')
    elif uses_directory:
        data_paths = []
        for f in os.listdir(directory):
            bool_list = [f.endswith(s) for s in possible_suf]
            if True in bool_list:
                data_paths.append(os.path.join(directory, f))
    # Read a zarr
    # -----------
    if is_zarr:
        if directory.endswith('.ome.zarr'):
            p = os.path.join(directory, '0')
            imgs = zarr.open(p)
        else:
            imgs = zarr.open(directory)
    # Everything that isnt a zarr 
    # ---------------------------
    else:
        imgs = []
        data_paths = sorted(data_paths)
        if in_memory:
            lazy_imread = None
        else:
            lazy_imread = delayed(imread)
        for p in data_paths:
            im = read_with_correct_modality(p, in_memory, lazy_imread)
            imgs.append(im)
        imgs = correct_shape(imgs)
        # Read in from a directory of files
        # ---------------------------------
        if uses_directory:
            if not in_memory:
                sample = imgs[0].compute
                [da.from_delayed(delayed_reader, shape=sample.shape, dtype=sample.dtype)
                for delayed_reader in imgs]
                del sample
            # if data is in stack/s (4D stacks of 3D frames)
            if data_type == 'image stacks' and len(imgs) > 1:
                if in_memory:
                    imgs = np.concatenate(imgs)
                else:
                    imgs = da.concatenate(imgs)
            # if data is in individual 3D frames
            if data_type == 'individual frames':
                if in_memory:
                    imgs = np.stack(imgs)
                else:
                    imgs = da.stack(imgs)
        # Single image - not zarr
        # -----------------------
        else:
            if in_memory:
                imgs = imgs[0]
            else:
                imgs = imgs[0].compute
                if imgs.ndim > 3:
                    chunks = (1, ) * imgs.ndim - 3 + imgs.shape[-3:]
                else:
                    chunks = 'auto'
                imgs = da.from_array(imgs, chunks=chunks)
    return imgs, uses_directory


def generate_4D_stack(path_list, shape):
    images = []
    for p in path_list:
        im  = read_with_correct_modality(p)
        assert shape == im.shape
        images.append(im)
    images = np.stack(images)
    assert len(images.shape) == 4
    return images


def read_with_correct_modality(path, in_memory, lazy_imread):
    if in_memory:
        if path.endswith('.tif') or path.endswith('.tiff'):
            im = imread(path)
        elif path.endswith('.ome.zarr'):
            p = os.path.join(path, '0')
            im = zarr.creation.open_array(p, 'r')
            im = np.array(im)
        elif path.endswith('.zar') or path.endswith('.zarr'):
            im = zarr.creation.open_array(path, 'r')
            im = np.array(im)
    else:
        if path.endswith('.tif') or path.endswith('.tiff'):
            im = lazy_imread(path)
    return im



def correct_shape(imgs):
    imgs = [np.squeeze(img) for img in imgs]
    shapes_3D = np.array([im.shape[-3:] for im in imgs])
    shape_3D = np.max(shapes_3D, axis=0)
    #max_size = [s == shape_3D for s in shapes_3D]
    not_max_size = [s != shape_3D for s in shapes_3D]
    if np.sum(not_max_size) > 0:
        shapes = np.array([im.shape for im in imgs])
        final_shapes = [list(s[:-3]) + list(shape_3D) for s in shapes]
        dtype = imgs[0].dtype
        final_imgs = [np.zeros(s, dtype=dtype) for s in final_shapes]
        slices = [tuple([slice(0, s) for s in shape]) for shape, im in zip(shapes, imgs)]
        for s_, img, orig in zip(slices, final_imgs, imgs):
            img[s_] = orig
        return final_imgs
    else:
        return imgs



# --------------
# Segment widget
# --------------

@magic_factory(
        save_dir={'widget_type': 'FileEdit', 'mode' : 'd'}, 
        chunk_size={'widget_type': 'LiteralEvalLineEdit'}, 
        margin={'widget_type': 'LiteralEvalLineEdit'}, 
        segmenter={'choices': list(segmenters.keys())}, 
        network_or_config_file={'widget_type': 'FileEdit'}, 
)
def segment_data(
    napari_viewer: napari.Viewer, 
    input_volume_layer: napari.layers.Image,
    save_dir: Union[str, None] = None,
    name: str = 'labels-prediction',
    segmenter: str='affinity-unet-watershed', 
    network_or_config_file: Union[str, None] = None,
    layer_reference: Union[str, None] = None,
    chunk_size: tuple = (10, 256, 256),
    margin: tuple = (1, 64, 64),
    debug: bool = True # TODO change to false
    ):
    '''
    Parameters
    ----------
    napari_viewer: napari.Viewer
        The napari viewer - this is automatically passed to magicgui.
    input_volume_layer: napari.layers.Image
        The image layer that you want to segment. 
    save_dir: str or None (None)
        The directory into which you want to save the data. If you
        do not choose one it will not be saved. Omit at own risk... 
    name: str ('labels-prediction')
        The name of the output layer and output file. 
    segmenter: str ('affinity-unet-watershed')
        This is the segmentation algorith you want to use. We have several 
        options:
            - affinity-unet-watershed: corresponds to the feature-based 
              segmentation trained using train_from_viewer. The unet predicts
              the object boundaries in each of three axes (affinities), 
              a mask (foreground), and the location of object centres. 
              These feature maps are used to segment the image with a
              modified watershed algorith.
            - unet-mask: segment with a unet to get only a single output
              channel with a mask. Will be added as an option for training
              in a future version. At the present this can be used to obtain
              only the mask from the affinity-unet-watershed. 
            - otsu-mask: a mask derived from smoothing the image with a 
              gaussian kernel then using otsu thresholding. To be added soon. 
            - blob-watershed: classical blob detection using a Laplacian of 
              Gaussian kernel to find object coordinate and inverse image intensity
              to segment the image using a cannonical watershed.
    network_or_config_file: str or None (None)
        A path to a neural network to use for the segmentation. If None and 
        using affinity-unet-watershed, the defaul U-net will be used. It is 
        trained to segment platelets. 
    layer_reference: str or None (None)
        A layer that the segmentation algorithm can get information from. 
        For instance, if the train_from_viewer function is used, it will 
        add a unet path to the layer metadata for any predicted labels.
        This can be accessed by affinity-unet-watershed. 
        TODO make sure this works!
    cchunk_size: tuple = (10, 256, 256) 
        What size of chunks would you like to compute scores for?  
        Each chunk is represented as a point on the VI and object count plots.
    margin: tuple = (1, 64, 64)
        This is the margin or overlap between chunks. The default is set this 
        way because this replicates the chunks that images might were segmented in. 
        The margin will change the number and position of chunks created. 
    debug: bool (False)
        Do you need to debug the code or save intermediate Unet output to check?
        If you use this, you will not be able to use the viewer until the code is 
        finished running or until it crashes. The reason for this is that the code 
        is run in the same thread as the viewer so  errors will be caught propery.
    '''
    seg_func = segmenters[segmenter]
    seg_func(napari_viewer, input_volume_layer, save_dir, 
             name, network_or_config_file, layer_reference, 
             chunk_size, margin, debug)


# --------------
# Combine Layers
# --------------

@magic_factory()
def combine_layers(
    napari_viewer: napari.Viewer, 
    base_layer: napari.layers.Layer, 
    to_append: napari.layers.Layer, 
    save_dir: Union[str, None] = None, 
    save_prefix: str = '', 
    save_all: bool = True, # save all or just new
    save_indivdually: bool =  False, 
    number_from: int = 0
):
    '''
    Combine a stack of labels volumes with a second stack. 

    Parameters
    ----------
    napari_viewer: napari.Viewer
    base_layer: napari.layers.Layer
        Layer to which the second layer will be added.
    to_append: napari.layers.Labels
        Layer to append to the base layer.
    save_dir: str or None
        If not none, the output will be saved to this 
        directory.
    save_prefix: str
        If save_dir is not None, this is the name that will 
        be used in saving the images.
    save_all: bool
        If save_dir is not None, should all of the images in 
        the stack be saved? If False, only the appended images
        will be saved
    save_individually: bool
        Should images be saved as a series of volumes or a single 
        4D stack?
    number_from: int
        if save_individually, images will be saved each with a unique 
        number, starting with this one (e.g., range starting from 0)
    '''
    base_layer.data = np.concatenate([base_layer.data, to_append.data])
    # what to do about metadata
    # should probably assert that the layer properties match

    # saving stuff... yay
    if save_dir is not None:
        if not save_all:
            if not save_indivdually:
                save_path = os.path.join(save_dir, save_prefix + '.zarr')
                zarr.save(save_path, to_append.data)
            else:
                for t in range(to_append.data.shape[0]):
                    save_path = os.path.join(save_dir, 
                                        save_prefix + f'_{t + number_from}.zarr')
                    zarr.save(save_path, to_append.data[t, ...])
        else:
            if not save_indivdually:
                save_path = os.path.join(save_dir, save_prefix + '.zarr')
                zarr.save(save_path, base_layer.data)
            else:
                for t in range(base_layer.data.shape[0]):
                    save_path = os.path.join(save_dir, 
                                        save_prefix + f'_{t + number_from}.zarr')
                    zarr.save(save_path, base_layer.data[t, ...])



# -----------------------
# Segmentation Assessment
# -----------------------
#@magic_factory(
#    data_path={'widget_type': 'FileEdit'}, 
#    data_type={'choices': ['individual frames', 'image stacks']},
#    layer_name={'widget_type' : 'LineEdit'},
#    layer_type={'choices': ['Image', 'Labels']},
#    scale={'widget_type' : 'LiteralEvalLineEdit'}, 
#    translate={'widget_type' : 'LiteralEvalLineEdit'}, 
#)

@magic_factory(
    save_dir={'widget_type': 'FileEdit', 'mode' : 'd'}, 
    chunk_size={'widget_type' : 'LiteralEvalLineEdit'}, 
    margin={'widget_type' : 'LiteralEvalLineEdit'},    
)
def assess_segmentation(
    napari_viewer: napari.Viewer,
    ground_truth: napari.layers.Labels, 
    model_segmentation: napari.layers.Labels, 
    chunk_size: tuple = (10, 256, 256), 
    margin: tuple = (1, 64, 64),
    variation_of_information: bool = True, 
    average_precision: bool = True, 
    object_count: bool = True, 
    #diagnostics: bool,
    save_dir: Union[str, None] = None, 
    save_prefix: str = 'segmentation-metrics',
    name: Union[str, None] = None, 
    show: bool = True, 
    exclude_chunks_less_than: int = 10,
    ):
    '''
    napari_viewer: napari.Viewer
        The napari viewer - this is automatically passed to magicgui.
    ground_truth: napari.layers.Labels
        The ground truth that you want to use to assess the segmentation.
    model_segmentation: napari.layers.Labels
        The segmentation you want to assess. 
    chunk_size: tuple = (10, 256, 256) 
        What size of chunks would you like to compute scores for?  
        Each chunk is represented as a point on the VI and object count plots.
    margin: tuple = (1, 64, 64)
        This is the margin or overlap between chunks. The default is set this 
        way because this replicates the chunks that images might were segmented in. 
        The margin will change the number and position of chunks created. 
    variation_of_information: (True)
        Will you assess VI?
        The variation of information is defined as VI(X,Y) = H(X|Y) + H(Y|X). 
        If X is the ground-truth segmentation, then H(X|Y) can be interpreted as 
        the amount of under-segmentation and H(Y|X) as the amount of 
        over-segmentation. In other words, a perfect over-segmentation will have 
        H(X|Y)=0 and a perfect under-segmentation will have H(Y|X)=0.
    average_precision: bool (True)
        Will you assess AP?
        Average precision is a combined measure of how accurate the model is at finding 
        true positive (real) objects (we call this precision) and how many of ground 
        truth real objects it found (this is called recall). 
            Precision = TP / (TP + FP)
            Recall = TP / (TP + FN)
            Abbreviations: FN, false negative; TP, true positive; FP, false positive 
        The assessment of whether an object is TP, FP, and FN depends on the threashold
        of overlap between objects. Here we use the intersection of union (IoU), which is
        the proportion of overlap between the bounding boxes of ground truth and model
        segemented objects. AP is assessed using different IoU thresholds (from 0.35-0.95). 
        The resultant data will be plotted as IoU by AP. 
    object_count: bool (True)
        Will you assess object count difference?
        The object count difference is simply the number of objects difference between the
        ground truth and the segementation. 
            card(ground truth) - card(segmentation)
        With this, the number of objects in ground truth and segementation + the percentage
        difference are added to the final data frame. 
    save_dir: str (None)
        This is the directory to which you will save the output. The assessment will not 
        run if you don't choose one. Ask yourself, why run an analysis without saving 
        anything? 
    save_prefix: str ('segmentation-metrics')
        This is the prefix that will be used to save the data. Several files will be saved:
            - <save_prefix>_AP_curve.csv (AP data)
            - <save_prefix>_scores.csv (VI, OC, and data used to create AP)
            - <save_prefix>_stats.csv (descriptive stats and 95% CI for variable is scores file)
    name: str (None)
        This is the name to give the model. This will be added in the data sheet in the column
        model_name. This is important because it is used in compare_segmentations to assign the 
        data to a group that can be compared to other groups. If None, the save_prefix will be
        used.   
    show: bool (True)
        Do you want to see the plots at the end?
    exclude_chunks_less_than: int (10)
        What is the minimum number of objects you need to see in a chunk of ground truth data
        to assess the chunk? Only use this if you KNOW you don't care about regions of the image 
        with few objects. You will only know this by visually inspecting the data + understanding
        your use case. 
    '''
    _assess_segmentation(ground_truth, model_segmentation, 
                         chunk_size, margin, variation_of_information, average_precision, 
                         object_count, save_dir, save_prefix, name, show, exclude_chunks_less_than)

    # Diagnostic plots
    # plot_diagnostics(...)
    #TODO:
    # - get image metrics relating to quality/information/randomness/edges
    # - get object metrics for segmentation 


def _assess_segmentation(
    ground_truth, 
    model_segmentation, 
    chunk_size: tuple = (10, 256, 256), 
    margin: tuple = (1, 64, 64),
    variation_of_information: bool = True, 
    average_precision: bool = True, 
    object_count: bool = True, 
    #diagnostics: bool,
    save_dir: str = 'choose directory', 
    save_prefix: str = 'segmentation-metrics',
    name: Union[str, None] = None,
    show: bool = True, 
    exclude_chunks_less_than: int = 10,
    ):
    if name == None:
        name = save_prefix
    # save info
    assert save_dir is not None, 'Please pick a directory to which to save the data.'
    os.makedirs(save_dir, exist_ok=True)
    data_path = os.path.join(save_dir, save_prefix + '_metrics.csv')
    # need to get the slices from the model-produced layer
    if isinstance(model_segmentation, napari.layers.Labels):
        shape = model_segmentation.data.shape
    else:
        shape = model_segmentation.shape
    slices = get_slices_from_chunks(shape, chunk_size, margin)
    data, stats = model_assessment(
        ground_truth, 
        model_segmentation, 
        save_prefix,
        name,
        slices,
        save_dir,
        variation_of_information, 
        average_precision, 
        object_count, 
        exclude_chunks_less_than)
    # generate plots
    plot_accuracy_metrics(
        data, 
        save_prefix, 
        save_dir, 
        name,
        variation_of_information, 
        average_precision, 
        object_count, 
        show
        )

    # Diagnostic plots
    # plot_diagnostics(...)
    #TODO:
    # - get image metrics relating to quality/information/randomness/edges
    # - get object metrics for segmentation 


def model_assessment(
    ground_truth: napari.layers.Labels, 
    model_segmentation: napari.layers.Labels, 
    save_prefix: str,
    name: str,
    slices: list,
    save_dir: str,
    variation_of_information: bool, 
    average_precision: bool, 
    object_count: bool, 
    exclude_chunks_less_than: int 
    ):
    # save info
    os.makedirs(save_dir, exist_ok=True)
    #data_path = os.path.join(save_dir, save_prefix + f'_{name}_metrics.csv')
    # need to get the slices from the model-produced layer
    data, stats = get_accuracy_metrics(slices, ground_truth, model_segmentation, name, save_prefix,
                                variation_of_information, average_precision, 
                                object_count, save_dir, exclude_chunks_less_than)
    return data, stats



def get_slices_from_chunks(arr_shape, chunk_size, margin):
    if len(arr_shape) <= 3:
        ts = range(1)
        fshape = arr_shape
    else:
        ts = range(arr_shape[0])
        fshape = arr_shape[1:]
    chunk_starts, chunk_crops = make_chunks(fshape, chunk_size, margin)
    slices = []
    for t in ts:
        for start, crop in list(zip(chunk_starts, chunk_crops)):
            sl = (slice(t, t+1), ) + tuple(
                    slice(start0, start0 + step)
                    for start0, step in zip(start, chunk_size)
                    )
            cr = tuple(slice(i, j) for i, j in crop)
            slices.append((sl, cr)) # useage: 4d_labels[sl][cr]
    return slices



# ---------------------------------
# Comparing models or ground truths
# ---------------------------------

@magic_factory(
    comparison_directory={'widget_type': 'FileEdit', 'mode' : 'd'}, 
    fig_size={'widget_type' : 'LiteralEvalLineEdit'}, 
    VI_indexs={'widget_type' : 'LiteralEvalLineEdit'},
    output_directory={'widget_type': 'FileEdit', 'mode' : 'd'}, 
    file_exstention={'choices': ['pdf', 'svg', 'png']},
)
def compare_segmentations(
       comparison_directory: str, 
        save_name: str,
        file_exstention: str ='pdf', 
        output_directory: Union[str, None] = None,
        variation_of_information: bool =True, 
        object_difference: bool =True, 
        average_precision: bool =True, 
        n_rows: int =2, 
        n_col: int =2, 
        comparison_name: str= "Model comparison",
        VI_indexs: tuple =(0, 1), # (0, 0)
        OD_index: int =2, # (0, 1)
        AP_index: int =3, # (1, 0)
        fig_size: tuple =(7, 6), 
        palette: str='Set2',
        top_white_space: float =5, #TODO eventually figure out how to make this a slider 0-100
        left_white_space: float =15, 
        right_white_space: float =5, 
        bottom_white_space: float =10, 
        horizontal_white_space: float =40,
        vertical_white_space: float =40,
        font_size: int =30,
        style: str ='ticks', 
        context: str='paper', 
        show: bool=True
    ):
    '''
    Make a custom plot to compare sementation models based on output from 
    the assess segmentation tool. The only requirement is that files you 
    want to compare are in the same directory, which should be specified
    as the comparison_directory. The plots will be saved into a single 
    file <save_name>.<file_exstention> (e.g., )

    Parameters
    ----------
    comparison_directory: str
        Directory in which to look for the data to plot
    save_name:
        Name to give the output file. Please don't add a file
        extension.
    file_exstention: str (.pdf)
        Specify one of the following file types: .pdf, .png, .svg
    output_directory: str or None (None)
        Directory into which to save the file. If None, 
        will save into the comparison directory. 
    variation_of_information: bool (True)
        Should we plot VI? This will be plotted in 2 plots:
        one for oversegmentation and one for undersegmentation. 
    object_difference: bool (True)
        Should we plot OD? Will be plotted into a single plot. 
    average_precision: bool (True)
        Should we plot AP? Will be plotted into a single plot.
    comparison_name: str ("Model comparison")
        Label to give comparison in plots. Will be used as
        an axis label in OD and VI plots
    n_rows: int (2) 
        How many rows of plots. 1 - 4.
        e.g. - two because we need all four plots for VI, OD, and AP
             - four because we need all four plots for VI, OD, and AP
               and want to plot everthing in the same column. 
    n_col: int =2, 
        How many rows of plots. 1 - 4.
        e.g. - two because we need all four plots for VI, OD, and AP
             - four because we need all four plots for VI, OD, and AP
               and want to plot everthing in the same row. 
        (e.g., two because we need all four plots for VI, OD, and AP)
    VI_indexs: tuple of int or int (0, 1)
        Which plot to put the VI plots in. The first index refers to
        the oversegmentation plot and the second refers to the 
        undersegmentation plot For instructions see "Using integer indexes" 
        in the function Notes below. 
    OD_index: tuple of int or int (2)
        Which plot to put the OD plot in.For instructions see 
        "Using integer indexes". 
    AP_index: tuple of int or int (3)
        Which plot to put the AP plot in.For instructions see 
        "Using integer indexes". 
    fig_size: tuple (9, 9)
        Size of the figure you want to make (in inches). 
    raincloud_orientation: str ("h")
        Orientation for raincloudplots. "h" for horizontal. 
        "v" for vertical. 
    raincloud_sigma: float (0.2)
        The sigma value used to construct the kernel density estimate 
        in the raincloudplot. Determines the smoothness of the 
        half violinplot/density estimate parts of the plot. Larger values
        result in smoother curves. 
    palette: str ('Set2')
        pandas colour palette to use. See the pandas palette documentation
        for details. 
    top_white_space: float (3)
        percent of the total figure size that should remain white space 
        above the plots.
    left_white_space: float (15)
        percent of the total figure size that should remain white space 
        to the left of the plots.
    right_white_space: float (5) 
        percent of the total figure size that should remain white space 
        to the right the plots.
    bottom_white_space: float (17)
        percent of the total figure size that should remain white space 
        below the plots.
    horizontal_white_space: float (16)
        percent of the total figure size that should remain white space 
        between the plots horizontally.
    vertical_white_space: float (16)
        percent of the total figure size that should remain white space 
        between the plots vertically.
    font_size: int (12)
        Size of the axis label and ticks font
    style: str ("ticks")
        Pandas style. Please see pandas documentation for more info and
        options.  
    context: str ("paper")
        Pandas context. Please see pandas documentation for more info and
        options.  
    show: bool (True)
        Do you want to see the plot in the matplotlib viewer once it has 
        been saved?

    Notes
    -----

        Using integer indexes
        ---------------------
        Numbers 0-4 that tells you which position to place the 
        oversegmentation and undersegmentation plots, respectively 
        Note that no matter how rows and columns are arranged, 
        the numbering will start at the top left plot and proceed
        left to write (much like reading English).
        e.g. - for VI plots, (0, 1) in a 2x2 grid of plots will place
               the VI plots in top two plots 
             - for OD plot, 3 in a 1x4 grid will place the OD plot 


    '''
    raincloud_orientation ='h'
    raincloud_sigma =0.2
    comparison_plots(comparison_directory, save_name, file_exstention, 
        output_directory, variation_of_information, 
        object_difference, average_precision, n_rows, n_col, 
        comparison_name, VI_indexs, OD_index, AP_index,
        fig_size, raincloud_orientation, raincloud_sigma,
        palette, top_white_space, left_white_space, 
        right_white_space, bottom_white_space, horizontal_white_space,
        vertical_white_space, font_size, style, context, show)


# -----------
# Save frames
# -----------

@magic_factory(
    save_dir={'widget_type': 'FileEdit', 'mode' : 'd'}, 
    frames={'widget_type' : 'LiteralEvalLineEdit'},
)
def save_frames(
        napari_viewer: napari.Viewer, 
        layer: napari.layers.Layer,
        save_dir: Union[str, None]=None,
        save_name: Union[str, None]=None,
        frames: Union[tuple, int]=None,
        save_as_stack: bool=True,
        load_saved: bool=False, 
        load_name: Union[str, None]=None,
    ):
    '''
    Save data from frames

    Parameters
    ----------
    napari_viewer: napari.Viewer
    layer: napari.layers.Layer
        The layer to save data from. Supported layers include:
        Image, Labels, Shapes, and Points. 
    save_dir: str
        The directory into which to save the data
    save_name:
        The name you want to use to save the data. This doesn't 
        need an extension. If a labels or image layer, data will
        be saved as a zarr file. If a shapes or points layer, 
        data will be saved as a npy file. 
    frames: tuple or None
        The frames you want to save (in brackets like so: (0, 2, 3))
        or None. If None, the whole layer will be saved. This will 
        only work for Labels and Image layers. All frames will always
        be saved for Points and Shapes layers. 
    
    '''
    #if isinstance() # check if dask array
    if isinstance(frames, int):
        frames = (frames, )
    if isinstance(layer, napari.layers.Image) or isinstance(layer, napari.layers.Labels):
        if isinstance(frames, tuple):
            slices = [slice(f, f+1) for f in frames]
            data = [layer.data[s] for s in slices]
            if save_as_stack:
                data = np.stack(data)
                data = np.squeeze(data)
                sp = os.path.join(str(save_dir), save_name + '.zarr')
                zarr.save(sp, data)
            else:
                for f, d in zip(frames, data):
                    sn = f'{save_name}_f{f}'
                    sp = os.path.join(str(save_dir), sn + '.zarr')
                    zarr.save(sp, d)
        if frames is None:
            if save_as_stack:
                sp = os.path.join(str(save_dir), save_name + '.zarr')
                data = np.squeeze(layer.data)
                zarr.save(sp, data)
            else:
                for f in layer.data.shape[0]:
                    sn = f'{save_name}_f{f}'
                    sp = os.path.join(str(save_dir), sn + '.zarr')
                    zarr.save(sp, layer.data[f, ...])
    elif isinstance(layer, napari.layers.Shapes):
        #TODO add shapes saving capabilities for ROIs
        data = layer.data
        data = np.stack(data)
        sp = os.path.join(str(save_dir), save_name + '.npy')
        np.save(sp, data, allow_pickle=False)
    elif isinstance(layer, napari.layers.Points):
        data = layer.data
        sp = os.path.join(str(save_dir), save_name + '.npy')
        np.save(sp, data, allow_pickle=False)
    load_saved_data(load_saved, napari_viewer, frames, layer, sp, load_name)
            


def load_saved_data(load_saved, napari_viewer, frames, layer, sp, load_name):
    if load_saved:
        if isinstance(layer, napari.layers.Image) or isinstance(layer, napari.layers.Labels):
            loaded = zarr.open(sp)
        elif isinstance(layer, napari.layers.Shapes) or isinstance(layer, napari.layers.Points):
            loaded = np.load(sp, allow_pickle=False)
        if load_name is None:
            fstr = '-'.join([str(f) for f in frames])
            load_name = f'{layer.name}_f{fstr}'
        if layer.ndim != loaded.ndim:
            diff = layer.ndim - loaded.ndim
            take_from = diff - layer.ndim
            scale = layer.scale[take_from:]
        else:
            scale = layer.scale
        if isinstance(layer, napari.layers.Image):
            napari_viewer.add_image(loaded, name=load_name, scale=scale)
        elif isinstance(layer, napari.layers.Labels):
            napari_viewer.add_labels(loaded, name=load_name, scale=scale)
        elif isinstance(layer, napari.layers.Shapes):
            napari_viewer.add_shapes(loaded, name=load_name, scale=scale)
        elif isinstance(layer, napari.layers.Points):
            napari_viewer.add_points(loaded, name=load_name, scale=scale)



# --------------------
# Generate GT from ROI
# --------------------

@magic_factory(
    save_dir={'widget_type': 'FileEdit', 'mode' : 'd'}, 
)
def ground_truth_from_ROI(
        napari_viewer: napari.Viewer, 
        image_layer: napari.layers.Image,
        labels_layer: napari.layers.Labels,
        shapes_layer: napari.layers.Shapes, 
        save_dir: Union[str, None]=None, 
        name: str='gt-from-ROI',
        number_of_tiles: int=1,
        padding: int=2
    ):
    '''
    Generate a ground truth image for training using
    several proofread ROIs.

    Parameters
    ----------
    napari_viewer: napari.Viewer
        The napari viewer. This is passed in automatically. 
    labels_layer: napari.layers.Labels
        The labels layer you want tp create ground truth from.
    shapes_layer: napari.layers.Shapes
        The shapes layer you want to use to find the ROIs. 
        At present, the shapes are treated as rectangles 
        across the xy plane. We will use the entire z axis.
    save_dir: str or None
        Folder to save data to. If this is not None, the data will
        be saved as a zarr file and the data will be opened into a 
        new layer. This means you can continue to proofread and this
        will be saved. 
    name: str='gt-from-ROI'
        Name that will be used for the output layer and for the save
        file. Note that this does not require a file extension. 
    number_of_tiles: int (1)
        The number of times each ROI will be repeated. Each will be saved 
        into its own frame and tiled as many times as specified. Note that
        if the number specifed is greater than the maximum times it fits
        into the frame, the number of tiles will be set to the maximum.
    padding: int (2)
        The number of pixels padding between each tile in the x and y axes. 
    '''
    _ground_truth_from_ROI(napari_viewer, image_layer, labels_layer, shapes_layer, 
                           save_dir, name, number_of_tiles, padding)


def _ground_truth_from_ROI(
        napari_viewer: napari.Viewer, 
        image_layer: napari.layers.Image, # TODO:save the image also with background noise
        labels_layer: napari.layers.Labels,
        shapes_layer: napari.layers.Shapes, 
        save_dir: Union[str, None]=None, 
        name: str='gt-from-ROI',
        number_of_tiles: int=1,
        padding: int=2
    ):
    '''
    Generate a ground truth image for training using
    several proofread ROIs.

    Parameters
    ----------
    napari_viewer: napari.Viewer
        The napari viewer. This is passed in automatically. 
    labels_layer: napari.layers.Labels
        The labels layer you want tp create ground truth from.
    shapes_layer: napari.layers.Shapes
        The shapes layer you want to use to find the ROIs. 
        At present, the shapes are treated as rectangles 
        across the xy plane. We will use the entire z axis.
    save_dir: str or None
        Folder to save data to. If this is not None, the data will
        be saved as a zarr file and the data will be opened into a 
        new layer. This means you can continue to proofread and this
        will be saved. 
    name: str='gt-from-ROI'
        Name that will be used for the output layer and for the save
        file. Note that this does not require a file extension. 
    number_of_tiles: int (1)
        The number of times each ROI will be repeated. Each will be saved 
        into its own frame and tiled as many times as specified. Note that
        if the number specifed is greater than the maximum times it fits
        into the frame, the number of tiles will be set to the maximum.
    padding: int (2)
        The number of pixels padding between each tile in the x and y axes. 
    '''
    gt = labels_layer.data
    img = image_layer.data
    rois = shapes_layer.data
    max_lim = [np.round(np.max(roi, axis=0)).astype(int) + 1 for roi in rois]
    min_lim = [np.round(np.min(roi, axis=0)).astype(int) for roi in rois]
    # Get the ROI labels
    # ------------------
    extra_dims = gt.ndim - 3
    slices = []
    for ll, ul in zip(min_lim, max_lim):
        s_xy = [slice(ll[i], ul[i]) for i in range(extra_dims + 1, gt.ndim)]
        s_z = [slice(None), ]
        s_extra = [slice(ll[i], ul[i]) for i in range(0, extra_dims)] * extra_dims 
        s = s_extra + s_z + s_xy
        s = tuple(s)
        slices.append(s)
    gt_rois = [gt[s_] for s_ in slices]
    im_rois = [img[s_] for s_ in slices]
    # Tile the ROIs as specified
    # --------------------------
    gt_xy_shape = gt.shape[-2:]
    max_per_axis = []
    max_per_axis = [np.floor_divide(gt_xy_shape, 
                        np.array(d.shape[-2:]) + padding).astype(int) \
                            for d in gt_rois]
    max_total = [mpa[0] * mpa[1] for mpa in max_per_axis]
    mod_per_axis = [np.mod(gt_xy_shape, 
                        (np.array(d.shape[-2:]) + padding)).astype(int) \
                            for d in gt_rois]
    final_gt_data = []
    final_im_data = []
    gt_3D_shape = gt.shape[-3:]
    for i, gt_roi in enumerate(gt_rois):
        mt = max_total[i]
        if number_of_tiles < mt:
            mt = number_of_tiles # we wont go past the max possible
        mapa = max_per_axis[i]
        y1 = gt_roi.shape[-2]
        x1 = gt_roi.shape[-1]
        slices = []
        for j in range(0, mt):
            # TODO add image and random background noise
            x_multi_factor = j % mapa[-1] + 1
            y_multi_factor = np.floor_divide(j, mapa[-1]) + 1
            px = (padding * x_multi_factor) - padding
            py = (padding * y_multi_factor) - padding
            _y1 = y1 * y_multi_factor + py
            _x1 = x1 * x_multi_factor + px
            y0 = _y1 - y1
            x0 = _x1 - x1
            s = (slice(None), slice(y0, _y1), slice(x0, _x1))
            slices.append(s)
        new_gt_frame = np.zeros(gt_3D_shape, dtype=gt.dtype)
        new_im_frame = np.random.normal(img.mean(), size=gt_3D_shape)
        #np.zeros(gt_3D_shape, dtype=img.dtype)
        for s in slices:
            new_gt_frame[s] = gt_roi
            new_im_frame[s] = im_rois[i]
        final_gt_data.append(new_gt_frame)
        final_im_data.append(new_im_frame)
    final_gt_data = np.stack(final_gt_data)
    final_gt_data = np.squeeze(final_gt_data)
    final_im_data = np.stack(final_im_data)
    final_im_data = np.squeeze(final_im_data)
    # Save if appropriate
    # -------------------
    if save_dir is not None:
        sp_l = os.path.join(save_dir, name + '_labels.zarr')
        zarr.save(sp_l, final_gt_data)
        sp_i = os.path.join(save_dir, name + '_img.zarr')
        zarr.save(sp_i, final_im_data)
        final_gt_data = zarr.open(sp_l)
        final_im_data = zarr.open(sp_i)
    # Add to viewer
    # -------------
    napari_viewer.add_image(final_im_data, scale=labels_layer.scale, 
                         translate=labels_layer.translate, name=name + '_img')
    napari_viewer.add_labels(final_gt_data, scale=labels_layer.scale, 
                         translate=labels_layer.translate, name=name + '_labels')
  

# ----------------
# Helper Functions
# ----------------

def find_matching_labels(
    napari_viewer: napari.Viewer, 
    labels
    ):
    # indices for the non background labels
    lab_idxs = np.where(labels > 0)
    matches = []
    for i, l in enumerate(napari_viewer.layers):
        if isinstance(l, napari.layers.labels.labels.Labels):
            res = np.min(l.data[lab_idxs] == labels[lab_idxs])
            if res == True:
                matches.append(i)
    if len(matches) > 1:
        print('multiple identical labels found... using the first...')
    return napari_viewer.layers[matches[0]]



